"""Tests for the Platform class."""

import copy
import io
import re
import warnings
from pathlib import Path
from queue import Queue
from types import MethodType
from unittest.mock import MagicMock, create_autospec, patch

import numpy as np
import pytest
from qibo import gates
from qibo.models import Circuit
from qpysequence import Sequence, Waveforms
from ruamel.yaml import YAML

from qililab import Arbitrary, save_platform
from qililab.constants import DEFAULT_PLATFORM_NAME
from qililab.digital import DigitalTranspilationConfig
from qililab.exceptions import ExceptionGroup
from qililab.instrument_controllers import InstrumentControllers
from qililab.instruments import SGS100A
from qililab.instruments.instruments import Instruments
from qililab.instruments.qblox import QbloxModule
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.platform import Bus, Buses, Platform
from qililab.pulse import Drag, Pulse, PulseEvent, PulseSchedule, Rectangular
from qililab.qprogram import Calibration, Domain, Experiment, QProgram
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.result.qblox_results import QbloxResult
from qililab.result.qprogram.qprogram_results import QProgramResults
from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult
from qililab.settings import AnalogCompilationSettings, DigitalCompilationSettings, Runcard
from qililab.settings.analog.flux_control_topology import FluxControlTopology
from qililab.settings.digital.gate_event_settings import GateEventSettings
from qililab.typings.enums import InstrumentName, Parameter
from qililab.waveforms import Chained, IQPair, Ramp, Square
from tests.data import Galadriel, SauronQDevil, SauronQuantumMachines, SauronSpiRack, SauronYokogawa
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform():
    return build_platform(runcard=Galadriel.runcard)


@pytest.fixture(name="platform_quantum_machines")
def fixture_platform_quantum_machines():
    return build_platform(runcard=SauronQuantumMachines.runcard)


@pytest.fixture(name="platform_spi")
def fixture_platform_spi():
    return build_platform(runcard=SauronSpiRack.runcard)


@pytest.fixture(name="platform_qdevil")
def fixture_platform_qdevil():
    return build_platform(runcard=SauronQDevil.runcard)


@pytest.fixture(name="platform_yokogawa")
def fixture_platform_yokogawa():
    return build_platform(runcard=SauronYokogawa.runcard)


@pytest.fixture(name="runcard")
def fixture_runcard():
    return Runcard(**copy.deepcopy(Galadriel.runcard))


@pytest.fixture(name="qblox_results")
def fixture_qblox_results():
    return [
        {
            "scope": {
                "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
            },
            "bins": {
                "integration": {"path0": [1], "path1": [1]},
                "threshold": [0],
                "avg_cnt": [1],
            },
            "qubit": 0,
            "measurement": 0,
        },
        {
            "scope": {
                "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
            },
            "bins": {
                "integration": {"path0": [1], "path1": [1]},
                "threshold": [1],
                "avg_cnt": [1],
            },
            "qubit": 0,
            "measurement": 1,
        },
        {
            "scope": {
                "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
            },
            "bins": {
                "integration": {"path0": [1], "path1": [1]},
                "threshold": [2],
                "avg_cnt": [1],
            },
            "qubit": 1,
            "measurement": 0,
        },
        {
            "scope": {
                "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
            },
            "bins": {
                "integration": {"path0": [1], "path1": [1]},
                "threshold": [3],
                "avg_cnt": [1],
            },
            "qubit": 1,
            "measurement": 1,
        },
    ]


@pytest.fixture(name="flux_to_bus_topology")
def get_flux_to_bus_topology():
    flux_control_topology_dict = [
        {"flux": "phix_q0", "bus": "flux_line_phix_q0"},
        {"flux": "phiz_q0", "bus": "flux_line_phiz_q0"},
        {"flux": "phix_q1", "bus": "flux_line_phix_q1"},
        {"flux": "phiz_q1", "bus": "flux_line_phiz_q1"},
        {"flux": "phix_c0_1", "bus": "flux_line_phix_c0_1"},
        {"flux": "phiz_c0_1", "bus": "flux_line_phiz_c0_1"},
    ]
    return [FluxControlTopology(**flux_control) for flux_control in flux_control_topology_dict]


@pytest.fixture(name="calibration")
def get_calibration():
    readout_duration = 2000
    readout_amplitude = 1.0
    r_wf_I = Square(amplitude=readout_amplitude, duration=readout_duration)
    r_wf_Q = Square(amplitude=0.0, duration=readout_duration)
    readout_waveform = IQPair(I=r_wf_I, Q=r_wf_Q)
    weights_shape = Square(amplitude=1, duration=readout_duration)
    weights = IQPair(I=weights_shape, Q=weights_shape)

    measurement_qp = QProgram()
    measurement_qp.measure(bus="readout_bus", waveform=readout_waveform, weights=weights)

    calibration = Calibration()
    calibration.add_block(name="measurement", block=measurement_qp.body)

    return calibration


@pytest.fixture(name="calibration_with_preparation_block")
def get_calibration_with_preparation_block():
    readout_duration = 2000
    readout_amplitude = 1.0
    r_wf_I = Square(amplitude=readout_amplitude, duration=readout_duration)
    r_wf_Q = Square(amplitude=0.0, duration=readout_duration)
    readout_waveform = IQPair(I=r_wf_I, Q=r_wf_Q)
    weights_shape = Square(amplitude=1, duration=readout_duration)
    weights = IQPair(I=weights_shape, Q=weights_shape)

    preparation_wf = Chained(
        waveforms=[Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=100), Square(amplitude=1.0, duration=200)]
    )
    preparation_qp = QProgram()
    preparation_qp.play(bus="flux_line_phix_q0", waveform=preparation_wf)
    preparation_qp.play(bus="flux_line_phiz_q0", waveform=preparation_wf)

    measurement_qp = QProgram()
    measurement_qp.measure(bus="readout_bus", waveform=readout_waveform, weights=weights)

    calibration = Calibration()
    calibration.add_block(name="preparation", block=preparation_qp.body)
    calibration.add_block(name="measurement", block=measurement_qp.body)

    return calibration


@pytest.fixture(name="anneal_qprogram")
def get_anneal_qprogram(runcard, flux_to_bus_topology):
    platform = Platform(runcard=runcard)
    platform.analog_compilation_settings = flux_to_bus_topology
    anneal_waveforms = {
        next(element.bus for element in platform.analog_compilation_settings if element.flux == "phix_q0"): Arbitrary(
            np.array([0.0, 0.0, 0.0, 1.0])
        ),
        next(element.bus for element in platform.analog_compilation_settings if element.flux == "phiz_q0"): Arbitrary(
            np.array([0.0, 0.0, 0.0, 2.0])
        ),
    }
    num_averages = 2
    num_shots = 1
    readout_duration = 2000
    readout_amplitude = 1.0
    r_wf_I = Square(amplitude=readout_amplitude, duration=readout_duration)
    r_wf_Q = Square(amplitude=0.0, duration=readout_duration)
    readout_waveform = IQPair(I=r_wf_I, Q=r_wf_Q)
    weights_shape = Square(amplitude=1, duration=readout_duration)
    weights = IQPair(I=weights_shape, Q=weights_shape)

    qp_anneal = QProgram()
    shots_variable = qp_anneal.variable("num_shots", Domain.Scalar, int)
    with qp_anneal.for_loop(variable=shots_variable, start=0, stop=num_shots, step=1):
        with qp_anneal.average(num_averages):
            for bus, waveform in anneal_waveforms.items():
                qp_anneal.play(bus=bus, waveform=waveform)
            qp_anneal.sync()
            with qp_anneal.block():
                qp_anneal.measure(bus="readout_bus", waveform=readout_waveform, weights=weights)
    return qp_anneal


@pytest.fixture(name="anneal_qprogram_with_preparation")
def get_anneal_qprogram_with_preparation(runcard, flux_to_bus_topology):
    platform = Platform(runcard=runcard)
    platform.analog_compilation_settings = flux_to_bus_topology
    anneal_waveforms = {
        next(element.bus for element in platform.analog_compilation_settings if element.flux == "phix_q0"): Arbitrary(
            np.array([0.0, 0.0, 0.0, 1.0])
        ),
        next(element.bus for element in platform.analog_compilation_settings if element.flux == "phiz_q0"): Arbitrary(
            np.array([0.0, 0.0, 0.0, 2.0])
        ),
    }
    num_averages = 2
    num_shots = 1
    readout_duration = 2000
    readout_amplitude = 1.0
    r_wf_I = Square(amplitude=readout_amplitude, duration=readout_duration)
    r_wf_Q = Square(amplitude=0.0, duration=readout_duration)
    readout_waveform = IQPair(I=r_wf_I, Q=r_wf_Q)
    weights_shape = Square(amplitude=1, duration=readout_duration)
    weights = IQPair(I=weights_shape, Q=weights_shape)
    preparation_wf = Chained(
        waveforms=[Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=100), Square(amplitude=1.0, duration=200)]
    )

    qp_anneal = QProgram()
    shots_variable = qp_anneal.variable("num_shots", Domain.Scalar, int)
    with qp_anneal.for_loop(variable=shots_variable, start=0, stop=num_shots, step=1):
        with qp_anneal.average(num_averages):
            with qp_anneal.block():
                qp_anneal.play(bus="flux_line_phix_q0", waveform=preparation_wf)
                qp_anneal.play(bus="flux_line_phiz_q0", waveform=preparation_wf)
            qp_anneal.sync()
            for bus, waveform in anneal_waveforms.items():
                qp_anneal.play(bus=bus, waveform=waveform)
            qp_anneal.sync()
            with qp_anneal.block():
                qp_anneal.measure(bus="readout_bus", waveform=readout_waveform, weights=weights)
    return qp_anneal


class TestPlatformInitialization:
    """Unit tests for the Platform class initialization"""

    def test_init_method(self, runcard):
        """Test initialization of the class"""
        platform = Platform(runcard=runcard)

        assert platform.name == runcard.name
        assert isinstance(platform.name, str)
        assert isinstance(platform.digital_compilation_settings, DigitalCompilationSettings)
        assert isinstance(platform.analog_compilation_settings, AnalogCompilationSettings)
        assert isinstance(platform.instruments, Instruments)
        assert isinstance(platform.instrument_controllers, InstrumentControllers)
        assert isinstance(platform.buses, Buses)
        assert platform._connected_to_instruments is False


class TestPlatform:
    """Unit tests checking the Platform class."""

    def test_platform_name(self, platform: Platform):
        """Test platform name."""
        assert platform.name == DEFAULT_PLATFORM_NAME

    def test_initial_setup_no_instrument_connection(self, platform: Platform):
        """Test platform raises and error if no instrument connection."""
        platform._connected_to_instruments = False
        with pytest.raises(
            AttributeError, match="Can not do initial_setup without being connected to the instruments."
        ):
            platform.initial_setup()

    @patch("qililab.typings.Parameter")
    def test_set_flux_parameter_qblox_channel0(self, mock_parameter, platform: Platform):
        """Test platform raises and error if no instrument connection."""
        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={"flux_line_q0_bus": {"flux_line_q0_bus": 0.1}}))
        platform.set_parameter(alias="flux_line_q0_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.OFFSET_OUT0.called_once()
        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={"flux_line_q1_bus": {"flux_line_q1_bus": 0.1}}))
        platform.set_parameter(alias="flux_line_q1_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.OFFSET_OUT1.called_once()
        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={"flux_line_q2_bus": {"flux_line_q2_bus": 0.1}}))
        platform.set_parameter(alias="flux_line_q2_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.OFFSET_OUT2.called_once()
        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={"flux_line_q3_bus": {"flux_line_q3_bus": 0.1}}))
        platform.set_parameter(alias="flux_line_q3_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.OFFSET_OUT3.called_once()

    @patch("qililab.typings.Parameter")
    def test_set_flux_parameter_spi(self, mock_parameter, platform_spi: Platform):
        """Test platform raises and error if no instrument connection."""
        platform_spi.set_crosstalk(CrosstalkMatrix.from_buses(buses={"spi_bus": {"spi_bus": 0.1}}))
        platform_spi.set_parameter(alias="spi_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.CURRENT.called_once()

    @patch("qililab.typings.Parameter")
    def test_set_flux_parameter_qdevil(self, mock_parameter, platform_qdevil: Platform):
        """Test platform raises and error if no instrument connection."""
        platform_qdevil.set_crosstalk(CrosstalkMatrix.from_buses(buses={"qdac_bus": {"qdac_bus": 0.1}}))
        platform_qdevil.set_parameter(alias="qdac_bus", parameter=Parameter.FLUX, value=0.14)
        assert mock_parameter.VOLTAGE.called_once()

    def test_set_flux_parameter_with_set_crosstalk(self, platform: Platform):
        """Test platform set FLUX parameter when crosstalk is given."""
        crosstalk_matrix = CrosstalkMatrix.from_buses(buses={"drive_line_q0_bus": {"drive_line_q0_bus": 0.1}})
        platform.set_crosstalk(crosstalk_matrix)
        platform.set_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX, value=0.14, channel_id=0)
        assert crosstalk_matrix == platform.crosstalk
        assert platform.get_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX, channel_id=0) == 0.14

    def test_set_flux_parameter_with_wrong_bus_raises_error(self, platform: Platform):
        """Test error raising when platform set FLUX alias is the wrong bus."""
        alias = "drive_line_q1_bus"
        crosstalk_matrix = CrosstalkMatrix.from_buses(buses={"drive_line_q0_bus": {"drive_line_q0_bus": 0.1}})
        error_string = f"{alias} not inside crosstalk matrix\n{crosstalk_matrix}"
        platform.set_crosstalk(crosstalk_matrix)
        with pytest.raises(ValueError, match=error_string):
            platform.set_parameter(alias=alias, parameter=Parameter.FLUX, value=0.14, channel_id=0)

    def test_set_flux_parameter_without_crosstalk_matrix_raises_error(self, platform: Platform):
        """Test error raised when the crostalk is not set"""
        error_string = "Crosstalk matrix has not been set"
        with pytest.raises(ValueError, match=error_string):
            platform.set_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX, value=0.14, channel_id=0)

    def test_set_flux_parameter_without_instruments_raises_error(self, platform_yokogawa: Platform):
        """Test error raised when the instruments do not match the flux parameter"""
        error_string = "Flux bus must have one of these instruments:\nQCM, QRM, QRM-RF, QCM-RF, D5a, S4g, quantum_machines_cluster, qdevil_qdac2"
        with pytest.raises(ReferenceError, match=error_string):
            crosstalk_matrix = CrosstalkMatrix.from_buses(
                buses={"yokogawa_gs200_current_bus": {"yokogawa_gs200_current_bus": 0.1}}
            )
            platform_yokogawa.set_crosstalk(crosstalk=crosstalk_matrix)
            platform_yokogawa.set_parameter(
                alias="yokogawa_gs200_current_bus",
                parameter=Parameter.FLUX,
                value=0.14,
                channel_id=0,
            )

    def test_set_flux_parameter_too_many_instruments_raises_error(self, platform: Platform):
        """Test error raised when there is more than one instrument affected by the flux"""
        error_string = "Flux bus must not have more than one of these instruments:\nQCM, QRM, QRM-RF, QCM-RF, D5a, S4g, quantum_machines_cluster, qdevil_qdac2"
        with pytest.raises(NotImplementedError, match=error_string):
            crosstalk_matrix = CrosstalkMatrix.from_buses(
                buses={"flux_line_too_many_instr": {"flux_line_too_many_instr": 0.1}}
            )
            platform.set_crosstalk(crosstalk=crosstalk_matrix)
            platform.set_parameter(
                alias="flux_line_too_many_instr",
                parameter=Parameter.FLUX,
                value=0.14,
                channel_id=0,
            )

    def test_set_flux_to_zero(self, platform: Platform):
        """Test set_flux_to_zero function."""
        crosstalk_matrix = CrosstalkMatrix.from_buses(buses={"drive_line_q0_bus": {"drive_line_q0_bus": 0.1}})
        platform.set_crosstalk(crosstalk_matrix)
        platform.set_flux_to_zero()
        assert platform.get_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX) == 0.0

    def test_set_flux_to_zero_without_crosstalk_raises_error(self, platform: Platform):
        """Test set_flux_to_zero function error without crosstalk."""
        error_string = "Crosstalk matrix has not been set"
        with pytest.raises(ValueError, match=error_string):
            platform.set_flux_to_zero()

    def test_set_bias_to_zero(self, platform: Platform):
        """Test set_bias_to_zero function."""
        crosstalk_matrix = CrosstalkMatrix.from_buses(buses={"drive_line_q0_bus": {"drive_line_q0_bus": 0.1}})
        platform.set_crosstalk(crosstalk_matrix)
        platform.set_bias_to_zero()
        assert platform.get_parameter(alias="drive_line_q0_bus", parameter=Parameter.OFFSET_OUT0) == 0.0

    def test_set_bias_to_zero_without_crosstalk_raises_error(self, platform: Platform):
        """Test set_bias_to_zero function error without crosstalk."""
        error_string = "Neither crosstalk matrix nor bus_list has been set"
        with pytest.raises(ValueError, match=error_string):
            platform.set_bias_to_zero()

    def test_set_parameter_no_instrument_connection_QBLOX(self, platform: Platform):
        """Test platform raises and error if no instrument connection."""
        platform._connected_to_instruments = False
        platform.set_parameter(alias="drive_line_q0_bus", parameter=Parameter.IF, value=0.14, channel_id=0)
        assert platform.get_parameter(alias="drive_line_q0_bus", parameter=Parameter.IF, channel_id=0) == 0.14

        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={"drive_line_q0_bus": {"drive_line_q0_bus": 0.1}}))
        platform.set_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX, value=0.14, channel_id=0)
        assert platform.get_parameter(alias="drive_line_q0_bus", parameter=Parameter.FLUX, channel_id=0) == 0.14

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("drive_q0_rf", Parameter.LO_FREQUENCY, 5e9),
            ("drive_q0_rf", Parameter.IF, 14e6),
            ("drive_q0_rf", Parameter.GAIN, 0.001),
            ("readout_q0_rf", Parameter.LO_FREQUENCY, 8e9),
            ("readout_q0_rf", Parameter.IF, 16e6),
            ("readout_q0_rf", Parameter.GAIN, 0.002),
            ("drive_q0", Parameter.IF, 13e6),
            ("flux_q0", Parameter.FLUX, 0.5),
        ],
    )
    def test_set_parameter_no_instrument_connection_QM(self, bus: str, parameter: Parameter, value: float | str | bool):
        """Test platform raises and error if no instrument connection."""
        # Overwrite platform to use Quantum Machines:
        platform = build_platform(runcard=SauronQuantumMachines.runcard)
        platform._connected_to_instruments = False

        platform.set_crosstalk(CrosstalkMatrix.from_buses(buses={bus: {bus: 1}}))
        platform.set_parameter(alias=bus, parameter=parameter, value=value)
        assert platform.get_parameter(alias=bus, parameter=parameter) == value

    def test_connect_logger(self, platform: Platform):
        platform._connected_to_instruments = True
        platform.instrument_controllers = MagicMock()
        with patch("qililab.platform.platform.logger", autospec=True) as mock_logger:
            platform.connect()
        mock_logger.info.assert_called_once_with("Already connected to the instruments")

    def test_disconnect_logger(self, platform: Platform):
        platform._connected_to_instruments = True
        platform.instrument_controllers = MagicMock()
        with patch("qililab.platform.platform.logger", autospec=True) as mock_logger:
            platform.disconnect()
        mock_logger.info.assert_called_once_with("Disconnected from instruments")

    def test_disconnect_fail_logger(self, platform: Platform):
        platform._connected_to_instruments = False
        platform.instrument_controllers = MagicMock()
        with patch("qililab.platform.platform.logger", autospec=True) as mock_logger:
            platform.disconnect()
        mock_logger.info.assert_called_once_with("Already disconnected from the instruments")

    def test_get_element_method_unknown_returns_none(self, platform: Platform):
        """Test get_element method with unknown element."""
        element = platform.get_element(alias="ABC")
        assert element is None

    def test_get_element_with_gate(self, platform: Platform):
        """Test the get_element method with a gate alias."""
        p_gates = platform.digital_compilation_settings.gates.keys()
        all(isinstance(event, GateEventSettings) for gate in p_gates for event in platform.get_element(alias=gate))

    def test_str_magic_method(self, platform: Platform):
        """Test __str__ magic method."""
        str(platform)

    def test_gates_settings_instance(self, platform: Platform):
        """Test settings instance."""
        assert isinstance(platform.digital_compilation_settings, DigitalCompilationSettings)

    def test_buses_instance(self, platform: Platform):
        """Test buses instance."""
        assert isinstance(platform.buses, Buses)

    def test_bus_0_signal_generator_instance(self, platform: Platform):
        """Test bus 0 signal generator instance."""
        element = platform.get_element(alias="rs_0")
        assert isinstance(element, SGS100A)

    def test_bus_1_awg_instance(self, platform: Platform):
        """Test bus 1 qubit readout instance."""
        element = platform.get_element(alias=f"{InstrumentName.QBLOX_QRM.value}_0")
        assert isinstance(element, QbloxModule)

    @patch("qililab.data_management.open")
    @patch("qililab.data_management.YAML.dump")
    def test_platform_manager_dump_method(self, mock_dump: MagicMock, mock_open: MagicMock, platform: Platform):
        """Test PlatformManager dump method."""
        save_platform(path="runcard.yml", platform=platform)
        mock_open.assert_called_once_with(file=Path("runcard.yml"), mode="w", encoding="utf-8")
        mock_dump.assert_called_once()

    @pytest.mark.parametrize("alias", ["drive_line_q0_bus", "drive_line_q1_bus", "feedline_input_output_bus", "foobar"])
    def test_get_bus_by_alias(self, platform: Platform, alias):
        """Test get_bus_by_alias method"""
        bus = platform._get_bus_by_alias(alias)
        if alias == "foobar":
            assert bus is None
        else:
            assert bus in platform.buses

    def test_print_platform(self, platform: Platform):
        """Test print platform."""
        assert str(platform) == str(YAML().dump(platform.to_dict(), io.BytesIO()))

    # I'm leaving this test here, because there is no test_instruments.py, but should be moved there when created
    def test_print_instruments(self, platform: Platform):
        """Test print instruments."""
        assert str(platform.instruments) == str(YAML().dump(platform.instruments._short_dict(), io.BytesIO()))

    def test_serialization(self, platform: Platform):
        """Test that a serialization of the Platform is possible"""
        runcard_dict = platform.to_dict()
        assert isinstance(runcard_dict, dict)

        new_platform = Platform(runcard=Runcard(**runcard_dict))
        assert isinstance(new_platform, Platform)
        assert str(new_platform) == str(platform)
        assert str(new_platform.name) == str(platform.name)
        assert str(new_platform.buses) == str(platform.buses)
        assert str(new_platform.instruments) == str(platform.instruments)
        assert str(new_platform.instrument_controllers) == str(platform.instrument_controllers)

        new_runcard_dict = new_platform.to_dict()
        assert isinstance(new_runcard_dict, dict)
        assert new_runcard_dict == runcard_dict

        newest_platform = Platform(runcard=Runcard(**new_runcard_dict))
        assert isinstance(newest_platform, Platform)
        assert str(newest_platform) == str(new_platform)
        assert str(newest_platform.name) == str(new_platform.name)
        assert str(newest_platform.buses) == str(new_platform.buses)
        assert str(newest_platform.instruments) == str(new_platform.instruments)
        assert str(newest_platform.instrument_controllers) == str(new_platform.instrument_controllers)


class TestMethods:
    """Unit tests for the methods of the Platform class."""

    def test_session_success(self):
        """Test the session method when everything works successfully."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the session method to the real one
        platform.session = Platform.session.__get__(platform, Platform)

        # Run the session successfully
        with platform.session():
            pass  # Simulate a successful experiment execution

        # Ensure methods were called in the correct order
        platform.connect.assert_called_once()
        platform.initial_setup.assert_called_once()
        platform.turn_on_instruments.assert_called_once()

        # Ensure cleanup is called in reverse order
        platform.turn_off_instruments.assert_called_once()
        platform.disconnect.assert_called_once()

    def test_session_with_exception(self):
        """Test the session method when an exception occurs during execution."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the session method to the real one
        platform.session = Platform.session.__get__(platform, Platform)

        # Simulate an exception during the experiment
        with pytest.raises(AttributeError, match="Test Error"):
            with platform.session():
                raise AttributeError("Test Error")

        # Ensure methods were called in the correct order before the exception
        platform.connect.assert_called_once()
        platform.initial_setup.assert_called_once()
        platform.turn_on_instruments.assert_called_once()

        # Ensure cleanup is still called in reverse order even after the exception
        platform.turn_off_instruments.assert_called_once()
        platform.disconnect.assert_called_once()

    def test_session_with_exception_in_setup(self):
        """Test the session method when an error occurs before turning on instruments."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the session method to the real one
        platform.session = Platform.session.__get__(platform, Platform)

        # Raise an exception after connect() and initial_setup() but before turn_on_instruments()
        platform.turn_on_instruments.side_effect = Exception("Instrument failure")

        # Simulate an error after connect() and initial_setup() but before turn_on_instruments()
        with pytest.raises(Exception, match="Instrument failure"):
            with platform.session():
                pass  # The exception will occur inside the context

        # Ensure methods were called until the point of failure
        platform.connect.assert_called_once()
        platform.initial_setup.assert_called_once()
        platform.turn_on_instruments.assert_called_once()

        # Ensure turn_off_instruments is not called, but disconnect is called
        platform.turn_off_instruments.assert_not_called()
        platform.disconnect.assert_called_once()

    def test_session_with_exception_in_cleanup(self):
        """Test the session method when an exception occurs during cleanup."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the session method to the real one
        platform.session = Platform.session.__get__(platform, Platform)

        # Simulate turn_off_instruments failing
        platform.turn_off_instruments.side_effect = Exception("Turn off instruments error")

        # Simulate no exception during the experiment, but failure during cleanup
        with pytest.raises(Exception, match="Turn off instruments error"):
            with platform.session():
                pass  # No exception during the experiment

        # Ensure methods were called in the correct order
        platform.connect.assert_called_once()
        platform.initial_setup.assert_called_once()
        platform.turn_on_instruments.assert_called_once()

        # Ensure the exception is raised in cleanup
        platform.turn_off_instruments.assert_called_once()
        platform.disconnect.assert_called_once()

    def test_session_with_multiple_exceptions_in_cleanup(self):
        """Test the session method when multiple exceptions occur during cleanup."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the session method to the real one
        platform.session = Platform.session.__get__(platform, Platform)

        # Simulate turn_off_instruments and disconnect failing
        platform.turn_off_instruments.side_effect = Exception("Turn off instruments error")
        platform.disconnect.side_effect = Exception("Disconnect error")

        with pytest.raises(ExceptionGroup) as exc_info:
            with platform.session():
                pass

        # Ensure ExceptionGroup has captured all exceptions
        assert len(exc_info.value.exceptions) == 2
        assert str(exc_info.value.exceptions[0]) == "Turn off instruments error"
        assert str(exc_info.value.exceptions[1]) == "Disconnect error"

        # Ensure methods were called in the correct order
        platform.connect.assert_called_once()
        platform.initial_setup.assert_called_once()
        platform.turn_on_instruments.assert_called_once()
        platform.turn_off_instruments.assert_called_once()
        platform.disconnect.assert_called_once()

    @pytest.mark.parametrize("optimize", [True, False])
    def test_compile_circuit(self, optimize: bool, platform: Platform):
        """Test the compilation of a qibo Circuit."""
        circuit = Circuit(3)
        circuit.add(gates.X(0))
        circuit.add(gates.X(1))
        circuit.add(gates.Y(0))
        circuit.add(gates.Y(1))
        circuit.add(gates.M(0, 1, 2))

        self._compile_and_assert(platform, circuit, 6, optimize=optimize)

    def test_compile_circuit_raises_error_if_digital_settings_missing(self, platform: Platform):
        """Test the compilation of a qibo Circuit."""
        circuit = Circuit(3)
        circuit.add(gates.X(0))
        circuit.add(gates.X(1))
        circuit.add(gates.Y(0))
        circuit.add(gates.Y(1))
        circuit.add(gates.M(0, 1, 2))

        platform.digital_compilation_settings = None

        with pytest.raises(ValueError):
            _ = platform.compile(program=circuit, num_avg=1000, repetition_duration=200_000, num_bins=1)

    def test_compile_pulse_schedule(self, platform: Platform):
        """Test the compilation of a qibo Circuit."""
        pulse_schedule = PulseSchedule()
        drag_pulse = Pulse(
            amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
        )
        readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
        pulse_schedule.add_event(PulseEvent(pulse=drag_pulse, start_time=0), bus_alias="drive_line_q0_bus", delay=0)
        pulse_schedule.add_event(
            PulseEvent(pulse=readout_pulse, start_time=200, qubit=0), bus_alias="feedline_input_output_bus", delay=0
        )

        self._compile_and_assert(platform, pulse_schedule, 2)

    def _compile_and_assert(
        self, platform: Platform, program: Circuit | PulseSchedule, len_sequences: int, optimize: bool = False
    ):
        sequences_w_alias, _ = platform.compile(
            program=program,
            num_avg=1000,
            repetition_duration=200_000,
            num_bins=1,
            transpilation_config=DigitalTranspilationConfig(optimize=optimize),
        )
        assert isinstance(sequences_w_alias, dict)
        if not optimize:
            assert len(sequences_w_alias) == len_sequences
        else:
            assert len(sequences_w_alias) < len_sequences
        for alias, sequences in sequences_w_alias.items():
            assert alias in {bus.alias for bus in platform.buses}
            assert isinstance(sequences, list)
            assert all(isinstance(sequence, Sequence) for sequence in sequences)
            assert sequences[0]._program.duration == 200_000 * 1000 + 4 + 4 + 4

    @pytest.mark.parametrize(
        "qprogram_fixture, calibration_fixture",
        [
            ("anneal_qprogram", "calibration"),
            ("anneal_qprogram_with_preparation", "calibration_with_preparation_block"),
        ],
    )
    def test_execute_anneal_program(
        self,
        platform: Platform,
        qprogram_fixture: str,
        flux_to_bus_topology: list[FluxControlTopology],
        calibration_fixture: str,
        request,
    ):
        anneal_qprogram = request.getfixturevalue(qprogram_fixture)
        calibration = request.getfixturevalue(calibration_fixture)

        mock_execute_qprogram = MagicMock()
        mock_execute_qprogram.return_value = QProgramResults()
        platform.execute_qprogram = mock_execute_qprogram  # type: ignore[method-assign]
        platform.analog_compilation_settings.flux_control_topology = flux_to_bus_topology
        transpiler = MagicMock()
        transpiler.return_value = (1, 2)

        results = platform.execute_annealing_program(
            annealing_program_dict=[{"qubit_0": {"sigma_x": 0.1, "sigma_z": 0.2}}],
            transpiler=transpiler,
            calibration=calibration,
            num_averages=2,
            num_shots=1,
        )
        qprogram = mock_execute_qprogram.call_args[1]["qprogram"].with_calibration(calibration)
        assert str(anneal_qprogram) == str(qprogram)
        assert isinstance(results, QProgramResults)

    def test_execute_anneal_program_no_measurement_raises_error(self, platform: Platform, calibration):
        mock_execute_qprogram = MagicMock()
        platform.execute_qprogram = mock_execute_qprogram  # type: ignore[method-assign]
        transpiler = MagicMock()
        transpiler.return_value = (1, 2)
        error_string = "The calibrated measurement is not present in the calibration file."
        with pytest.raises(ValueError, match=error_string):
            platform.execute_annealing_program(
                annealing_program_dict=[{"qubit_0": {"sigma_x": 0.1, "sigma_z": 0.2}}],
                transpiler=transpiler,
                calibration=calibration,
                num_averages=2,
                num_shots=1,
                measurement_block="whatever",
            )

    def test_execute_experiment(self):
        """Test the execute_experiment method of the Platform class."""
        # Create an autospec of the Platform class
        platform = create_autospec(Platform, instance=True)

        # Manually set the execute_experiment method to the real one
        platform.execute_experiment = MethodType(Platform.execute_experiment, platform)

        # Create an autospec of the Experiment class and Calibration class
        mock_experiment = create_autospec(Experiment)

        expected_results_path = "mock/results/path/data.h5"

        # Mock the ExperimentExecutor to ensure it's used correctly
        with patch("qililab.platform.platform.ExperimentExecutor") as MockExecutor:
            mock_executor_instance = MockExecutor.return_value  # Mock instance of ExperimentExecutor
            mock_executor_instance.execute.return_value = expected_results_path

            # Call the method under test
            results_path = platform.execute_experiment(experiment=mock_experiment)

            # Check that ExperimentExecutor was instantiated with the correct arguments
            MockExecutor.assert_called_once_with(platform=platform, experiment=mock_experiment)

            # Ensure the execute method was called on the ExperimentExecutor instance
            mock_executor_instance.execute.assert_called_once()

            # Ensure that execute_experiment returns the correct value
            assert results_path == expected_results_path

    def test_execute_qprogram_with_qblox(self, platform: Platform):
        """Test that the execute method compiles the qprogram, calls the buses to run and return the results."""
        drive_wf = IQPair(I=Square(amplitude=1.0, duration=40), Q=Square(amplitude=0.0, duration=40))
        readout_wf = IQPair(I=Square(amplitude=1.0, duration=120), Q=Square(amplitude=0.0, duration=120))
        weights_wf = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        qprogram = QProgram()
        qprogram.play(bus="drive_line_q0_bus", waveform=drive_wf)
        qprogram.play(bus="drive_line_q1_bus", waveform=drive_wf)
        qprogram.sync()
        qprogram.play(bus="feedline_input_output_bus", waveform=readout_wf)
        qprogram.play(bus="feedline_input_output_bus_1", waveform=readout_wf)
        qprogram.qblox.acquire(bus="feedline_input_output_bus", weights=weights_wf)
        qprogram.qblox.acquire(bus="feedline_input_output_bus_1", weights=weights_wf)

        with (
            patch("builtins.open") as patched_open,
            patch.object(Bus, "upload_qpysequence") as upload,
            patch.object(Bus, "run") as run,
            patch.object(Bus, "acquire_qprogram_results") as acquire_qprogram_results,
            patch.object(QbloxModule, "sync_sequencer") as sync_sequencer,
            patch.object(QbloxModule, "desync_sequencer") as desync_sequencer,
        ):
            acquire_qprogram_results.return_value = [123]
            first_execution_results = platform.execute_qprogram(qprogram=qprogram)

            acquire_qprogram_results.return_value = [456]
            second_execution_results = platform.execute_qprogram(qprogram=qprogram)

            _ = platform.execute_qprogram(qprogram=qprogram, debug=True)

        # assert upload executed only once (2 because there are 2 buses)
        assert upload.call_count == 4

        # assert run executed all three times (6 because there are 2 buses)
        assert run.call_count == 12
        assert acquire_qprogram_results.call_count == 6  # only readout buses
        assert sync_sequencer.call_count == 12  # called as many times as run
        assert desync_sequencer.call_count == 12
        assert first_execution_results.results["feedline_input_output_bus"] == [123]
        assert first_execution_results.results["feedline_input_output_bus_1"] == [123]
        assert second_execution_results.results["feedline_input_output_bus"] == [456]
        assert second_execution_results.results["feedline_input_output_bus_1"] == [456]

        # assure only one debug was called
        assert patched_open.call_count == 1

    def test_execute_qprogram_with_qblox_distortions(self, platform: Platform):
        drive_wf = Square(amplitude=1.0, duration=4)
        qprogram = QProgram()
        qprogram.play(bus="drive_line_q0_bus", waveform=drive_wf)

        test_waveforms_q0 = Waveforms()
        test_waveforms_q0.add(array=[0.5, 1.0, 0.5, 0.0], index=0)
        test_waveforms_q0.add(array=[0.0, 0.0, 0.0, 0.0], index=1)

        with (
            patch("builtins.open"),
            patch.object(Bus, "upload_qpysequence") as upload,
            patch.object(Bus, "run"),
            patch.object(Bus, "acquire_qprogram_results"),
            patch.object(QbloxModule, "sync_sequencer"),
            patch.object(QbloxModule, "desync_sequencer"),
        ):
            _ = platform.execute_qprogram(qprogram=qprogram)
            assert test_waveforms_q0.to_dict() == upload.call_args_list[0].kwargs["qpysequence"]._waveforms.to_dict()

    def test_execute_qprogram_with_quantum_machines(self, platform_quantum_machines: Platform):
        """Test that the execute_qprogram method executes the qprogram for Quantum Machines correctly"""
        drive_wf = IQPair(I=Square(amplitude=1.0, duration=40), Q=Square(amplitude=0.0, duration=40))
        readout_wf = IQPair(I=Square(amplitude=1.0, duration=120), Q=Square(amplitude=0.0, duration=120))
        weights_wf = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        qprogram = QProgram()
        qprogram.play(bus="drive_q0_rf", waveform=drive_wf)
        qprogram.sync()
        qprogram.play(bus="readout_q0_rf", waveform=readout_wf)
        qprogram.measure(bus="readout_q0_rf", waveform=readout_wf, weights=weights_wf)

        with (
            patch("builtins.open") as patched_open,
            patch("qililab.platform.platform.generate_qua_script", return_value=None) as generate_qua,
            patch.object(QuantumMachinesCluster, "config") as config,
            patch.object(QuantumMachinesCluster, "append_configuration") as append_configuration,
            patch.object(QuantumMachinesCluster, "compile") as compile_program,
            patch.object(QuantumMachinesCluster, "run_compiled_program") as run_compiled_program,
            patch.object(QuantumMachinesCluster, "get_acquisitions") as get_acquisitions,
        ):
            cluster = platform_quantum_machines.get_element("qmm")
            config.return_value = cluster.settings.to_qua_config()

            get_acquisitions.return_value = {"I_0": np.array([1, 2, 3]), "Q_0": np.array([4, 5, 6])}
            first_execution_results = platform_quantum_machines.execute_qprogram(qprogram=qprogram)

            get_acquisitions.return_value = {"I_0": np.array([3, 2, 1]), "Q_0": np.array([6, 5, 4])}
            second_execution_results = platform_quantum_machines.execute_qprogram(qprogram=qprogram)

            _ = platform_quantum_machines.execute_qprogram(qprogram=qprogram, debug=True)

        assert compile_program.call_count == 3
        assert append_configuration.call_count == 3
        assert run_compiled_program.call_count == 3
        assert get_acquisitions.call_count == 3

        assert "readout_q0_rf" in first_execution_results.results
        assert len(first_execution_results.results["readout_q0_rf"]) == 1
        assert isinstance(first_execution_results.results["readout_q0_rf"][0], QuantumMachinesMeasurementResult)
        np.testing.assert_array_equal(first_execution_results.results["readout_q0_rf"][0].I, np.array([1, 2, 3]))
        np.testing.assert_array_equal(first_execution_results.results["readout_q0_rf"][0].Q, np.array([4, 5, 6]))

        assert "readout_q0_rf" in second_execution_results.results
        assert len(second_execution_results.results["readout_q0_rf"]) == 1
        assert isinstance(second_execution_results.results["readout_q0_rf"][0], QuantumMachinesMeasurementResult)
        np.testing.assert_array_equal(second_execution_results.results["readout_q0_rf"][0].I, np.array([3, 2, 1]))
        np.testing.assert_array_equal(second_execution_results.results["readout_q0_rf"][0].Q, np.array([6, 5, 4]))

        # assure only one debug was called
        assert patched_open.call_count == 1
        assert generate_qua.call_count == 1

    def test_execute_qprogram_with_quantum_machines_raises_error(self, platform_quantum_machines: Platform):
        """Test that the execute_qprogram method raises the exception if the qprogram failes"""

        error_string = "The QM `config` dictionary does not exist. Please run `initial_setup()` first."
        escaped_error_str = re.escape(error_string)
        platform_quantum_machines.compile = MagicMock()  # type: ignore # don't care about compilation
        platform_quantum_machines.compile.return_value = Exception(escaped_error_str)

        drive_wf = IQPair(I=Square(amplitude=1.0, duration=40), Q=Square(amplitude=0.0, duration=40))
        readout_wf = IQPair(I=Square(amplitude=1.0, duration=120), Q=Square(amplitude=0.0, duration=120))
        weights_wf = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        qprogram = QProgram()
        qprogram.play(bus="drive_q0_rf", waveform=drive_wf)
        qprogram.sync()
        qprogram.play(bus="readout_q0_rf", waveform=readout_wf)
        qprogram.measure(bus="readout_q0_rf", waveform=readout_wf, weights=weights_wf)

        with patch.object(QuantumMachinesCluster, "turn_off") as turn_off:
            with pytest.raises(ValueError, match=escaped_error_str):
                _ = platform_quantum_machines.execute_qprogram(qprogram=qprogram, debug=True)

        turn_off.assert_called_once_with()

    def test_execute(self, platform: Platform, qblox_results: list[dict]):
        """Test that the execute method calls the buses to run and return the results."""
        # Define pulse schedule
        pulse_schedule = PulseSchedule()
        drag_pulse = Pulse(
            amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
        )
        readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
        pulse_schedule.add_event(PulseEvent(pulse=drag_pulse, start_time=0), bus_alias="drive_line_q0_bus", delay=0)
        pulse_schedule.add_event(
            PulseEvent(pulse=readout_pulse, start_time=200, qubit=0), bus_alias="feedline_input_output_bus", delay=0
        )
        qblox_result = QbloxResult(qblox_raw_results=qblox_results, integration_lengths=[1, 1, 1, 1])
        with patch.object(Bus, "upload") as upload:
            with patch.object(Bus, "run") as run:
                with patch.object(Bus, "acquire_result") as acquire_result:
                    with patch.object(QbloxModule, "desync_sequencers") as desync:
                        acquire_result.return_value = qblox_result
                        result = platform.execute(
                            program=pulse_schedule, num_avg=1000, repetition_duration=2000, num_bins=1
                        )

        assert upload.call_count == len(pulse_schedule.elements)
        assert run.call_count == len(pulse_schedule.elements)
        acquire_result.assert_called_once_with()
        assert result == qblox_result
        desync.assert_called()

    def test_execute_with_queue(self, platform: Platform, qblox_results: list[dict]):
        """Test that the execute method adds the obtained results to the given queue."""
        queue: Queue = Queue()
        pulse_schedule = PulseSchedule()
        pulse_schedule.add_event(
            PulseEvent(
                pulse=Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular()),
                start_time=200,
                qubit=0,
            ),
            bus_alias="feedline_input_output_bus",
            delay=0,
        )
        qblox_result = QbloxResult(qblox_raw_results=qblox_results, integration_lengths=[1, 1, 1, 1])
        with patch.object(Bus, "upload"):
            with patch.object(Bus, "run"):
                with patch.object(Bus, "acquire_result") as acquire_result:
                    with patch.object(QbloxModule, "desync_sequencers") as desync:
                        acquire_result.return_value = qblox_result
                        _ = platform.execute(
                            program=pulse_schedule, num_avg=1000, repetition_duration=2000, num_bins=1, queue=queue
                        )

        assert len(queue.queue) == 1
        assert queue.get() == qblox_result
        desync.assert_called()

    def test_execute_returns_ordered_measurements(self, platform: Platform, qblox_results: list[dict]):
        """Test that executing with some circuit returns acquisitions with multiple measurements in same order
        as they appear in circuit"""

        # Define circuit
        c = Circuit(2)
        c.add([gates.M(1), gates.M(0), gates.M(0, 1)])  # without ordering, these are retrieved for each sequencer, so
        # the order from qblox qrm will be M(0),M(0),M(1),M(1)

        for idx, final_layout in enumerate([{0: 0, 1: 1}, {0: 1, 1: 0}]):
            platform.compile = MagicMock()  # type: ignore # don't care about compilation
            platform.compile.return_value = {"feedline_input_output_bus": None}, final_layout
            with patch.object(Bus, "upload"):
                with patch.object(Bus, "run"):
                    with patch.object(Bus, "acquire_result") as acquire_result:
                        with patch.object(QbloxModule, "desync_sequencers"):
                            acquire_result.return_value = QbloxResult(
                                qblox_raw_results=qblox_results, integration_lengths=[1, 1, 1, 1]
                            )
                            result = platform.execute(program=c, num_avg=1000, repetition_duration=2000, num_bins=1)

            # check that the order of #measurement # qubit is the same as in the circuit
            order_measurement_qubit = [(result["qubit"], result["measurement"]) for result in result.qblox_raw_results]  # type: ignore

            # Change the qubit mappings, given the final_layout:
            assert (
                order_measurement_qubit == [(1, 0), (0, 0), (0, 1), (1, 1)]
                if idx == 0
                else [(0, 0), (1, 0), (1, 1), (0, 1)]
            )

    def test_execute_no_readout_raises_error(self, platform: Platform, qblox_results: list[dict]):
        """Test that executing with some circuit returns acquisitions with multiple measurements in same order
        as they appear in circuit"""

        # Define circuit
        c = Circuit(2)
        c.add([gates.M(1), gates.M(0), gates.M(0, 1)])  # without ordering, these are retrieved for each sequencer, so
        # the order from qblox qrm will be M(0),M(0),M(1),M(1)

        # compile will return nothing and thus
        # readout_buses = [
        #     bus for bus in self.buses if isinstance(bus.system_control, ReadoutSystemControl) and bus.alias in programs
        # ]
        # in platform will be empty
        platform.compile = MagicMock()  # type: ignore # don't care about compilation
        platform.compile.return_value = {"drive_line_q0_bus": None}, {"q0": 0}
        with patch.object(Bus, "upload"):
            with patch.object(Bus, "run"):
                with patch.object(Bus, "acquire_result") as acquire_result:
                    with patch.object(QbloxModule, "desync_sequencers"):
                        acquire_result.return_value = QbloxResult(
                            qblox_raw_results=qblox_results, integration_lengths=[1, 1, 1, 1]
                        )
                        error_string = "There are no readout buses in the platform."
                        with pytest.raises(ValueError, match=error_string):
                            _ = platform.execute(program=c, num_avg=1000, repetition_duration=20_000, num_bins=1)

    def test_order_results_circuit_M_neq_acquisitions(self, platform: Platform, qblox_results: list[dict]):
        """Test that executing with some circuit returns acquisitions with multiple measurements in same order
        as they appear in circuit"""

        # Define circuit
        c = Circuit(2)
        c.add([gates.M(1), gates.M(0, 1)])  # without ordering, these are retrieved for each sequencer, so
        # the order from qblox qrm will be M(0),M(1),M(1)
        n_m = len([qubit for gate in c.queue for qubit in gate.qubits if isinstance(gate, gates.M)])

        platform.compile = MagicMock()  # type: ignore[method-assign] # don't care about compilation
        platform.compile.return_value = {"feedline_input_output_bus": None}, {"q0": 0}
        with patch.object(Bus, "upload"):
            with patch.object(Bus, "run"):
                with patch.object(Bus, "acquire_result") as acquire_result:
                    with patch.object(QbloxModule, "desync_sequencers"):
                        acquire_result.return_value = QbloxResult(
                            qblox_raw_results=qblox_results, integration_lengths=[1, 1, 1, 1]
                        )
                        error_string = f"Number of measurements in the circuit {n_m} does not match number of acquisitions {len(qblox_results)}"
                        with pytest.raises(ValueError, match=error_string):
                            _ = platform.execute(program=c, num_avg=1000, repetition_duration=2000, num_bins=1)

    def test_execute_raises_error_if_program_type_wrong(self, platform: Platform):
        """Test that `Platform.execute` raises an error if the program sent is not a Circuit or a PulseSchedule."""
        c = Circuit(1)
        c.add(gates.M(0))
        program = [c, c]
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Program to execute can only be either a single circuit or a pulse schedule. Got program of type {type(program)} instead"
            ),
        ):
            platform.execute(program=program, num_avg=1000, repetition_duration=2000, num_bins=1)

    def test_execute_stack_2qrm(self, platform: Platform):
        """Test that the execute stacks results when more than one qrm is called."""
        # build mock qblox results
        qblox_raw_results = QbloxResult(
            integration_lengths=[20],
            qblox_raw_results=[
                {
                    "scope": {
                        "path0": {"data": [1, 1, 1, 1, 1, 1, 1, 1], "out-of-range": False, "avg_cnt": 1000},
                        "path1": {"data": [0, 0, 0, 0, 0, 0, 0, 0], "out-of-range": False, "avg_cnt": 1000},
                    },
                    "bins": {
                        "integration": {"path0": [1, 1, 1, 1], "path1": [0, 0, 0, 0]},
                        "threshold": [0.5, 0.5, 0.5, 0.5],
                        "avg_cnt": [1000, 1000, 1000, 1000],
                    },
                    "qubit": 0,
                    "measurement": 0,
                }
            ],
        )
        pulse_schedule = PulseSchedule()
        # mock compile method
        platform.compile = MagicMock()  # type: ignore[method-assign]
        platform.compile.return_value = (
            {"feedline_input_output_bus": None, "feedline_input_output_bus_2": None},
            {"q0": 0},
        )
        # mock execution
        with patch.object(Bus, "upload"):
            with patch.object(Bus, "run"):
                with patch.object(Bus, "acquire_result") as acquire_result:
                    with patch.object(QbloxModule, "desync_sequencers"):
                        acquire_result.return_value = qblox_raw_results
                        result = platform.execute(
                            program=pulse_schedule, num_avg=1000, repetition_duration=2000, num_bins=1
                        )
        assert len(result.qblox_raw_results) == 2  # type: ignore[attr-defined]
        assert qblox_raw_results.qblox_raw_results[0] == result.qblox_raw_results[0]  # type: ignore[attr-defined]
        assert qblox_raw_results.qblox_raw_results[0] == result.qblox_raw_results[1]  # type: ignore[attr-defined]

    @pytest.mark.parametrize("parameter", [Parameter.AMPLITUDE, Parameter.DURATION, Parameter.PHASE])
    @pytest.mark.parametrize("gate", ["I(0)", "X(0)", "Y(0)"])
    @pytest.mark.parametrize("value", [1.0, 100, 0.0])
    def test_set_parameter_of_gates(self, parameter, gate, value, platform: Platform):
        """Test the ``get_parameter`` method with gates."""
        platform.set_parameter(parameter=parameter, alias=gate, value=value)
        gate_settings = platform.digital_compilation_settings.gates[gate][0]
        assert getattr(gate_settings.pulse, parameter.value) == value

        platform.digital_compilation_settings = None
        with pytest.raises(ValueError):
            platform.set_parameter(parameter=parameter, alias=gate, value=value)

    @pytest.mark.parametrize("parameter", [Parameter.AMPLITUDE, Parameter.DURATION, Parameter.PHASE])
    @pytest.mark.parametrize("gate", ["I(0)", "X(0)", "Y(0)"])
    def test_get_parameter_of_gates(self, parameter, gate, platform: Platform):
        """Test the ``get_parameter`` method with gates."""
        gate_settings = platform.digital_compilation_settings.gates[gate][0]
        assert platform.get_parameter(parameter=parameter, alias=gate) == getattr(gate_settings.pulse, parameter.value)

        platform.digital_compilation_settings = None
        with pytest.raises(ValueError):
            platform.get_parameter(parameter=parameter, alias=gate)

    @pytest.mark.parametrize("parameter", [Parameter.DRAG_COEFFICIENT, Parameter.NUM_SIGMAS])
    @pytest.mark.parametrize("gate", ["X(0)", "Y(0)"])
    def test_get_parameter_of_pulse_shapes(self, parameter, gate, platform: Platform):
        """Test the ``get_parameter`` method with gates."""
        gate_settings = platform.digital_compilation_settings.gates[gate][0]
        assert platform.get_parameter(parameter=parameter, alias=gate) == gate_settings.pulse.shape[parameter.value]

    def test_get_parameter_of_gates_raises_error(self, platform: Platform):
        """Test that the ``get_parameter`` method with gates raises an error when a gate is not found."""
        with pytest.raises(KeyError, match="Gate Drag for qubits 3 not found in settings"):
            platform.get_parameter(parameter=Parameter.AMPLITUDE, alias="Drag(3)")

    @pytest.mark.parametrize("parameter", [Parameter.DELAY_BEFORE_READOUT])
    def test_get_parameter_of_platform(self, parameter, platform: Platform):
        """Test the ``get_parameter`` method with platform parameters."""
        value = platform.get_parameter(parameter=parameter, alias="platform")
        assert value == 0

    @pytest.mark.parametrize("parameter", [Parameter.FLUX])
    def test_get_flux_parameter(self, parameter, platform: Platform):
        """Test the ``get_parameter`` method with platform flux parameters. The default as 0 is created"""
        value = platform.get_parameter(parameter=parameter, alias="flux_line_phix_q0")
        assert value == 0

    def test_get_parameter_with_delay(self, platform: Platform):
        """Test the ``get_parameter`` method with the delay of a bus."""
        value = platform.get_parameter(parameter=Parameter.DELAY, alias="drive_line_q0_bus")
        assert value == 0

    @pytest.mark.parametrize(
        "parameter",
        [Parameter.IF, Parameter.GAIN, Parameter.LO_FREQUENCY, Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT2],
    )
    def test_get_parameter_of_bus(self, parameter, platform: Platform):
        """Test the ``get_parameter`` method with the parameters of a bus."""
        CHANNEL_ID = 0
        bus = platform._get_bus_by_alias(alias="drive_line_q0_bus")
        assert bus is not None
        assert bus.get_parameter(parameter=parameter, channel_id=CHANNEL_ID) == platform.get_parameter(
            parameter=parameter, alias="drive_line_q0_bus", channel_id=CHANNEL_ID
        )

    def test_no_bus_to_flux_raises_error(self, platform: Platform):
        """Test that if flux to bus topology is not specified an error is raised"""
        platform.analog_compilation_settings = None
        error_string = "Flux to bus topology not given in the runcard"
        with pytest.raises(ValueError, match=error_string):
            platform.execute_annealing_program(
                annealing_program_dict=[{}],
                calibration=MagicMock(),
                transpiler=MagicMock(),
                num_averages=2,
                num_shots=1,
            )

    def test_get_element_flux(self, platform: Platform):
        """Get the bus from a flux using get_element"""
        for flux in ["phiz_q0", "phix_c0_1"]:
            bus = platform.get_element(flux)
            assert bus.alias == next(
                flux_bus.bus
                for flux_bus in platform.analog_compilation_settings.flux_control_topology
                if flux_bus.flux == flux
            )

    def test_parallelisation_same_bus_raises_error_qblox(self, platform: Platform):
        """Test that if parallelisation is attempted on qprograms using at least one bus in common, an error will be raised"""
        error_string = "QPrograms cannot be executed in parallel."
        qp1 = QProgram()
        qp2 = QProgram()
        qp3 = QProgram()
        qp1.play(bus="drive_line_q0_bus", waveform=Square(amplitude=1, duration=5))
        qp2.play(bus="drive_line_q1_bus", waveform=Square(amplitude=1, duration=25))
        qp2.play(bus="drive_line_q0_bus", waveform=Square(amplitude=0.5, duration=35))
        qp3.play(bus="feedline_input_output_bus_1", waveform=Square(amplitude=0.5, duration=15))

        with (
            patch("builtins.open") as patched_open,
            patch.object(Bus, "upload_qpysequence") as upload,
            patch.object(Bus, "run") as run,
            patch.object(Bus, "acquire_qprogram_results") as acquire_qprogram_results,
            patch.object(QbloxModule, "sync_sequencer") as sync_sequencer,
            patch.object(QbloxModule, "desync_sequencer") as desync_sequencer,
        ):
            qp_list = [qp1, qp2, qp3]
            with pytest.raises(ValueError, match=error_string):
                platform.execute_qprograms_parallel(qp_list, debug=True)

    def test_parallelisation_execute_quantum_machine_not_supported(self, platform_quantum_machines: Platform):
        error_string = "Parallel execution is not supported in Quantum Machines."
        qp1 = QProgram()
        qp2 = QProgram()
        qp3 = QProgram()
        qp1.play(bus="drive_q0", waveform=Square(amplitude=1, duration=5))
        qp2.play(bus="flux_q0", waveform=Square(amplitude=1, duration=25))

        with (
            patch("builtins.open") as patched_open,
            patch.object(Bus, "upload_qpysequence") as upload,
            patch.object(Bus, "run") as run,
            patch.object(Bus, "acquire_qprogram_results") as acquire_qprogram_results,
            patch.object(QbloxModule, "sync_sequencer") as sync_sequencer,
            patch.object(QbloxModule, "desync_sequencer") as desync_sequencer,
        ):
            qp_list = [qp1, qp2, qp3]
            with pytest.raises(ValueError, match=error_string):
                platform_quantum_machines.execute_qprograms_parallel(qp_list)

    def test_parallelisation_execute_qblox(self, platform: Platform):
        """Test that the execute parallelisation returns the same result per qprogram as the regular excute method"""

        drive_wf = IQPair(I=Square(amplitude=1.0, duration=40), Q=Square(amplitude=0.0, duration=40))
        readout_wf = IQPair(I=Square(amplitude=1.0, duration=120), Q=Square(amplitude=0.0, duration=120))
        weights_wf = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        drive_wf2 = IQPair(I=Square(amplitude=1.0, duration=80), Q=Square(amplitude=0.0, duration=80))
        readout_wf2 = IQPair(I=Square(amplitude=1.0, duration=150), Q=Square(amplitude=0.0, duration=150))
        weights_wf2 = IQPair(I=Square(amplitude=1.0, duration=2200), Q=Square(amplitude=0.0, duration=2200))

        qprogram1 = QProgram()
        qprogram1.play(bus="drive_line_q0_bus", waveform=drive_wf)
        qprogram1.play(bus="drive_line_q1_bus", waveform=drive_wf)
        qprogram1.sync()
        qprogram1.play(bus="feedline_input_output_bus", waveform=readout_wf)
        qprogram1.play(bus="feedline_input_output_bus_1", waveform=readout_wf)
        qprogram1.qblox.acquire(bus="feedline_input_output_bus", weights=weights_wf)

        qprogram2 = QProgram()
        qprogram2.play(bus="flux_line_q0_bus", waveform=drive_wf2)
        qprogram2.sync()
        qprogram2.play(bus="feedline_input_output_bus_2", waveform=readout_wf2)
        qprogram2.qblox.acquire(bus="feedline_input_output_bus_2", weights=weights_wf2)

        with (
            patch("builtins.open") as patched_open,
            patch.object(Bus, "upload_qpysequence") as upload,
            patch.object(Bus, "run") as run,
            patch.object(Bus, "acquire_qprogram_results") as acquire_qprogram_results,
            patch.object(QbloxModule, "sync_sequencer") as sync_sequencer,
            patch.object(QbloxModule, "desync_sequencer") as desync_sequencer,
        ):
            acquire_qprogram_results.return_value = [123]
            qp_list = [qprogram1, qprogram2]
            result_parallel = platform.execute_qprograms_parallel(qp_list, debug=True)
            non_parallel_results1 = platform.execute_qprogram(qprogram=qprogram1, debug=True)
            non_parallel_results2 = platform.execute_qprogram(qprogram=qprogram2, debug=True)

            # check that each element of the result list of the parallel execution is the same as the regular execution for each respective qprograms
            assert result_parallel[0].results == non_parallel_results1.results
            assert result_parallel[1].results == non_parallel_results2.results

    def test_calibrate_mixers(self, platform: Platform):
        """Test calibrating the Qblox mixers."""
        channel_id = 0
        cal_type = "lo"
        alias_drive_bus = "drive_line_q1_bus"
        alias_readout_bus = "feedline_input_output_bus_1"
        drive_bus = platform.get_element(alias=alias_drive_bus)
        readout_bus = platform.get_element(alias=alias_readout_bus)
        qcm_rf = drive_bus.instruments[0]
        qrm_rf = readout_bus.instruments[0]

        qcm_rf.calibrate_mixers = MagicMock()
        qrm_rf.calibrate_mixers = MagicMock()

        platform.calibrate_mixers(alias=alias_drive_bus, cal_type=cal_type, channel_id=channel_id)
        qcm_rf.calibrate_mixers.assert_called_with(cal_type, channel_id)

        platform.calibrate_mixers(alias=alias_readout_bus, cal_type=cal_type, channel_id=channel_id)
        qrm_rf.calibrate_mixers.assert_called_with(cal_type, channel_id)

        cal_type = "lo_and_sidebands"

        platform.calibrate_mixers(alias=alias_drive_bus, cal_type=cal_type, channel_id=channel_id)
        qcm_rf.calibrate_mixers.assert_called_with(cal_type, channel_id)

        platform.calibrate_mixers(alias=alias_readout_bus, cal_type=cal_type, channel_id=channel_id)
        qrm_rf.calibrate_mixers.assert_called_with(cal_type, channel_id)

        cal_type = "lo"
        non_rf_drive_bus = "drive_line_q0_bus"
        non_rf_readout_bus = "feedline_input_output_bus"

        with pytest.raises(AttributeError, match="Mixers calibration not implemented for this instrument."):
            platform.calibrate_mixers(alias=non_rf_drive_bus, cal_type=cal_type, channel_id=channel_id)

        with pytest.raises(AttributeError, match="Mixers calibration not implemented for this instrument."):
            platform.calibrate_mixers(alias=non_rf_readout_bus, cal_type=cal_type, channel_id=channel_id)

        cal_type = "lo_and_sidebands"

        with pytest.raises(AttributeError, match="Mixers calibration not implemented for this instrument."):
            platform.calibrate_mixers(alias=non_rf_drive_bus, cal_type=cal_type, channel_id=channel_id)

        with pytest.raises(AttributeError, match="Mixers calibration not implemented for this instrument."):
            platform.calibrate_mixers(alias=non_rf_readout_bus, cal_type=cal_type, channel_id=channel_id)



""" Tests for Dictionaries utilities"""

import pytest

from qililab.utils.dictionaries import merge_dictionaries


@pytest.fixture(name="origin")
def fixture_origin_dictionary() -> dict:
    origin = {
        "should_not_update": {"a": 1, "b": 2, "c": 3},
        "should_update_inner_element_c": {"a": 1, "b": 2, "c": 3},
        "should_update": {},
    }

    return origin


@pytest.fixture(name="new")
def fixture_new_dictionary() -> dict:
    new = {"should_update_inner_element_c": {"c": 100}, "should_update": {"a": 1, "b": 2, "c": 3}}

    return new


class TestDictionaries:
    """Unit tests for utils.dictionaries module"""

    def test_merge_dictionaries(self, origin: dict, new: dict):
        merged = merge_dictionaries(origin, new)

        assert len(merged["should_not_update"]) == 3
        assert merged["should_not_update"]["a"] == 1
        assert merged["should_not_update"]["b"] == 2
        assert merged["should_not_update"]["c"] == 3

        assert len(merged["should_update_inner_element_c"]) == 3
        assert merged["should_update_inner_element_c"]["a"] == 1
        assert merged["should_update_inner_element_c"]["b"] == 2
        assert merged["should_update_inner_element_c"]["c"] == 100

        assert len(merged["should_update"]) == 3
        assert merged["should_update"]["a"] == 1
        assert merged["should_update"]["b"] == 2
        assert merged["should_update"]["c"] == 3



"""Test for the VectorNetworkAnalyzer E5071B class."""
import copy
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from qililab.instrument_controllers.vector_network_analyzer.agilent_E5071B_vna_controller import E5071BController
from qililab.instruments.agilent.e5071b_vna import E5071B
from qililab.instruments.instrument import ParameterNotFound
from qililab.platform import Platform
from qililab.result.vna_result import VNAResult
from qililab.typings.enums import Parameter, VNAScatteringParameters, VNATriggerModes
from tests.data import SauronVNA
from tests.test_utils import build_platform


@pytest.fixture(name="sauron_platform")
def fixture_sauron_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronVNA.runcard)


@pytest.fixture(name="e5071b_controller")
def fixture_e5071b_controller(sauron_platform: Platform):
    """Return an instance of VectorNetworkAnalyzer controller class"""
    settings = copy.deepcopy(SauronVNA.agilent_e5071b_controller)
    settings.pop("name")
    return E5071BController(settings=settings, loaded_instruments=sauron_platform.instruments)


@pytest.fixture(name="e5071b_no_device")
def fixture_e5071b_no_device():
    """Return an instance of VectorNetworkAnalyzer class"""
    settings = copy.deepcopy(SauronVNA.agilent_e5071b)
    settings.pop("name")
    return E5071B(settings=settings)


@pytest.fixture(name="e5071b")
@patch(
    "qililab.instrument_controllers.vector_network_analyzer.agilent_E5071B_vna_controller.VectorNetworkAnalyzerDriver",
    autospec=True,
)
def fixture_e5071b(mock_device: MagicMock, e5071b_controller: E5071BController):
    """Return connected instance of VectorNetworkAnalyzer class"""
    mock_instance = mock_device.return_value
    e5071b_controller.connect()
    mock_device.assert_called()
    return e5071b_controller.modules[0]


class TestE5071B:
    """Unit tests checking the VectorNetworkAnalyzer E5071B attributes and methods"""

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.POWER, -60.0),
            (Parameter.IF_BANDWIDTH, 1.5),
            (Parameter.FREQUENCY_SPAN, 6.4e-3),
            (Parameter.FREQUENCY_CENTER, 8.5e-3),
            (Parameter.FREQUENCY_START, 27.5),
            (Parameter.FREQUENCY_STOP, 40.5),
            (Parameter.ELECTRICAL_DELAY, 0.0),
        ],
    )
    def test_setup_method_value_flt(self, parameter: Parameter, value, e5071b: E5071B, e5071b_no_device: E5071B):
        """Test the setup method with float value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, float)
        for e5071bs in [e5071b, e5071b_no_device]:
            e5071bs.setup(parameter, value)
            if parameter == Parameter.POWER:
                assert e5071bs.settings.power == value
            if parameter == Parameter.FREQUENCY_SPAN:
                assert e5071bs.frequency_span == value
            if parameter == Parameter.FREQUENCY_CENTER:
                assert e5071bs.frequency_center == value
            if parameter == Parameter.FREQUENCY_START:
                assert e5071bs.frequency_start == value
            if parameter == Parameter.FREQUENCY_STOP:
                assert e5071bs.frequency_stop == value
            if parameter == Parameter.ELECTRICAL_DELAY:
                assert e5071bs.electrical_delay == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, 0.34),
            (Parameter.VOLTAGE, -20.1),
        ],
    )
    def test_setup_method_flt_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect float parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.TRIGGER_MODE, "INT"),
            (Parameter.SCATTERING_PARAMETER, "S21"),
        ],
    )
    def test_setup_method_value_str(self, parameter: Parameter, value, e5071b: E5071B):
        """Test the setup method with str value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, str)
        e5071b.setup(parameter, value)
        if parameter == Parameter.TRIGGER_MODE:
            assert e5071b.trigger_mode == VNATriggerModes(value)
        if parameter == Parameter.SCATTERING_PARAMETER:
            assert e5071b.scattering_parameter == VNAScatteringParameters(value)

    @pytest.mark.parametrize(
        "value",
        [
            "S221",
            "s11",
        ],
    )
    def test_setup_scattering_value_raises_exception(self, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect str value"""
        assert isinstance(value, str)
        with pytest.raises(ValueError):
            e5071b.scattering_parameter = value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, "foo"),
            (Parameter.VOLTAGE, "bar"),
        ],
    )
    def test_setup_method_str_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect str parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value", [(Parameter.AVERAGING_ENABLED, True), (Parameter.AVERAGING_ENABLED, False)]
    )
    def test_setup_method_value_bool(self, parameter: Parameter, value, e5071b: E5071B):
        """Test the setup method with bool value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, bool)
        e5071b.setup(parameter, value)
        if parameter == Parameter.AVERAGING_ENABLED:
            assert e5071b.averaging_enabled == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, True),
            (Parameter.VOLTAGE, False),
        ],
    )
    def test_setup_method_bool_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect bool parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.setup(parameter, value)

    @pytest.mark.parametrize("parameter, value", [(Parameter.NUMBER_POINTS, 100), (Parameter.NUMBER_AVERAGES, 4)])
    def test_setup_method_value_int(self, parameter: Parameter, value, e5071b: E5071B):
        """Test the setup method with int value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, int)
        e5071b.setup(parameter, value)
        if parameter == Parameter.NUMBER_POINTS:
            assert e5071b.number_points == value
        if parameter == Parameter.NUMBER_AVERAGES:
            assert e5071b.number_averages == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Test POWER setting
            (Parameter.POWER, 2.0),

            # Test IF_BANDWIDTH setting
            (Parameter.IF_BANDWIDTH, 1.0),

            # Test ELECTRICAL_DELAY setting
            (Parameter.ELECTRICAL_DELAY, 1.0)
        ]
    )
    def test_set_parameter(self, e5071b: E5071B, parameter, value):
        """Test setting parameters for E5071B functionality using parameterized values."""
        e5071b.set_parameter(parameter, value, channel_id=1)

        # Check values based on the parameter
        if parameter == Parameter.POWER:
            assert e5071b.settings.power == value
        elif parameter == Parameter.IF_BANDWIDTH:
            assert e5071b.settings.if_bandwidth == value
        elif parameter == Parameter.ELECTRICAL_DELAY:
            assert e5071b.settings.electrical_delay == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.GATE_DURATION, 0),
        ],
    )
    def test_set_parameter_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with not supported parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.set_parameter(parameter, value)

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            # Test POWER setting
            (Parameter.POWER, -60.0),
            # Test IF_BANDWITH setting
            (Parameter.IF_BANDWIDTH, 1.5),
            # Test ELECTRICAL_DELAY setting
            (Parameter.ELECTRICAL_DELAY, 0.0)
        ]
    )
    def test_get_parameter(self, e5071b: E5071B, parameter, expected_value):
        """Test setting parameters for E5071B functionality using parameterized values."""
        if parameter == Parameter.IF_BANDWIDTH:
            e5071b.set_parameter(Parameter.IF_BANDWIDTH, 1.5)
        value = e5071b.get_parameter(parameter, channel_id=1)
        assert value == expected_value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.GATE_DURATION, 0),
        ],
    )
    def test_get_parameter_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with not supported parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.get_parameter(parameter)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, 0),
            (Parameter.VOLTAGE, -20),
        ],
    )
    def test_setup_method_int_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect int parameter"""
        with pytest.raises(ParameterNotFound):
            e5071b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.SCATTERING_PARAMETER, ["S221"]),
            (Parameter.SCATTERING_PARAMETER, {}),
        ],
    )
    def test_setup_method_raises_exception(self, parameter, value, e5071b: E5071B):
        """Test the setup method raises exception with incorrect value type"""
        with pytest.raises(ParameterNotFound):
            e5071b.setup(parameter, value)

    def test_to_dict_method(self, e5071b_no_device: E5071B):
        """Test the dict method"""
        assert isinstance(e5071b_no_device.to_dict(), dict)

    def test_reset_method(self, e5071b: E5071B):
        """Test the reset method"""
        e5071b.reset()
        e5071b.device.reset.assert_called()

    def test_turn_on_method(self, e5071b: E5071B):
        """Test turn on method"""
        e5071b.turn_on()
        e5071b.device.send_command.assert_called_with(command=":OUTP", arg="ON")

    def test_turn_off_method(self, e5071b: E5071B):
        """Test turn off method"""
        e5071b.turn_off()
        e5071b.device.send_command.assert_called_with(command=":OUTP", arg="OFF")

    @pytest.mark.parametrize("command, arg", [(":SENS1:AVER:CLE", ""), ("SENS1:AVER:COUN", "3")])
    def test_send_command_method(self, command, arg, e5071b: E5071B):
        """Test the send command method"""
        assert isinstance(command, str)
        assert isinstance(arg, str)
        e5071b.send_command(command, arg)
        e5071b.device.send_command.assert_called_with(command, arg)

    @patch("numpy.frombuffer")
    def test_get_data_method(self, mock_frombuffer, e5071b: E5071B):  # sourcery skip: simplify-division
        """Test the get data method"""
        mock_buffer = MagicMock(name="mock_frombuffer")
        mock_frombuffer.return_value = mock_buffer
        e5071b.get_data()
        e5071b.device.send_command.assert_called()
        e5071b.device.read_raw.assert_called()

    @patch("numpy.frombuffer")
    def test_acquire_result_method(self, mock_frombuffer, e5071b: E5071B):
        """Test the acquire result method"""
        mock_buffer = MagicMock(name="mock_frombuffer")
        mock_frombuffer.return_value = mock_buffer
        output = e5071b.acquire_result()
        assert isinstance(output, VNAResult)
        assert np.allclose(output.array, [])
        assert output.to_dict() == {}

    @pytest.mark.parametrize("continuous", [True, False])
    def test_continuous_method(self, continuous, e5071b: E5071B):
        """Test the continuous method"""
        e5071b.continuous(continuous)
        if continuous:
            e5071b.device.send_command.assert_called_with(":INIT:CONT", "ON")
        else:
            e5071b.device.send_command.assert_called_with(":INIT:CONT", "OFF")

    @pytest.mark.parametrize("query", [":SENS1:SWE:MODE?"])
    def test_send_query_method(self, query: str, e5071b: E5071B):
        """Test the send query method"""
        assert isinstance(query, str)
        e5071b.send_query(query)
        e5071b.device.send_query.assert_called_with(query)

    @pytest.mark.parametrize("query", ["SENS:X?", "CALC1:MEAS1:DATA:SDAT?"])
    def test_send_binary_query_method(self, query: str, e5071b: E5071B):
        """Test the send binary query method"""
        assert isinstance(query, str)
        e5071b.send_binary_query(query)
        e5071b.device.send_binary_query.assert_called_with(query)

    def test_read_method(self, e5071b: E5071B):
        """Test the read method"""
        e5071b.read()
        e5071b.device.read.assert_called()

    def test_read_raw_method(self, e5071b: E5071B):
        """Test the read method"""
        e5071b.read_raw()
        e5071b.device.read_raw.assert_called()

    def test_set_timeout_method(self, e5071b: E5071B):
        """Test the set timeout method"""
        e5071b.set_timeout(100)
        e5071b.device.set_timeout.assert_called_with(100)

    @pytest.mark.parametrize(
        "state, command, arg", [(True, "SENS1:AVER:STAT", "ON"), (False, "SENS1:AVER:STAT", "OFF")]
    )
    def test_average_state_method(self, state, command, arg, e5071b: E5071B):
        """Test the auxiliary private method average state"""
        e5071b._average_state(state)
        e5071b.device.send_command.assert_called_with(command=command, arg=arg)

    def test_average_count_method(self, e5071b: E5071B):
        """Set the average count"""
        e5071b._average_count(1, 1)  # Have to know if you can select which one was called
        e5071b.device.send_command.assert_called()

    def test_autoscale_method(self, e5071b: E5071B):
        """Test the autoscale method"""
        e5071b.autoscale()
        e5071b.device.send_command.assert_called_with(command="DISP:WIND:TRAC:Y:AUTO", arg="")

    def test_scattering_parameter_property(self, e5071b_no_device: E5071B):
        """Test the scattering parametter property"""
        assert hasattr(e5071b_no_device, "scattering_parameter")
        assert e5071b_no_device.scattering_parameter == e5071b_no_device.settings.scattering_parameter

    def test_frequency_span_property(self, e5071b_no_device: E5071B):
        """Test the frequency span property"""
        assert hasattr(e5071b_no_device, "frequency_span")
        assert e5071b_no_device.frequency_span == e5071b_no_device.settings.frequency_span

    def test_frequency_center_property(self, e5071b_no_device: E5071B):
        """Test the frequency center property"""
        assert hasattr(e5071b_no_device, "frequency_center")
        assert e5071b_no_device.frequency_center == e5071b_no_device.settings.frequency_center

    def test_frequency_start_property(self, e5071b_no_device: E5071B):
        """Test the frequency start property"""
        assert hasattr(e5071b_no_device, "frequency_start")
        assert e5071b_no_device.frequency_start == e5071b_no_device.settings.frequency_start

    def test_frequency_stop_property(self, e5071b_no_device: E5071B):
        """Test the frequency stop property"""
        assert hasattr(e5071b_no_device, "frequency_stop")
        assert e5071b_no_device.frequency_stop == e5071b_no_device.settings.frequency_stop

    def test_averaging_enabled_property(self, e5071b_no_device: E5071B):
        """Test the averaging enabled property"""
        assert hasattr(e5071b_no_device, "averaging_enabled")
        assert e5071b_no_device.averaging_enabled == e5071b_no_device.settings.averaging_enabled

    def test_number_averages_propertyy(self, e5071b_no_device: E5071B):
        """Test the number of averages property"""
        assert hasattr(e5071b_no_device, "number_averages")
        assert e5071b_no_device.number_averages == e5071b_no_device.settings.number_averages

    def test_trigger_mode_property(self, e5071b_no_device: E5071B):
        """Test the trigger mode property"""
        assert hasattr(e5071b_no_device, "trigger_mode")
        assert e5071b_no_device.trigger_mode == e5071b_no_device.settings.trigger_mode

    def test_number_points_property(self, e5071b_no_device: E5071B):
        """Test the number points property"""
        assert hasattr(e5071b_no_device, "number_points")
        assert e5071b_no_device.number_points == e5071b_no_device.settings.number_points

    def test_electrical_delay_property(self, e5071b_no_device: E5071B):
        """Test the electrical delay property"""
        assert hasattr(e5071b_no_device, "electrical_delay")
        assert e5071b_no_device.electrical_delay == e5071b_no_device.settings.electrical_delay


import re
from unittest.mock import MagicMock, call

import pytest

from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.qdevil.qdevil_qdac2 import QDevilQDac2
from qililab.typings.enums import Parameter
from qililab.waveforms import Square


@pytest.fixture(name="qdac")
def fixture_qdac() -> QDevilQDac2:
    """Fixture that returns an instance of a dummy QDAC-II."""
    qdac = QDevilQDac2(
        {
            "alias": "qdac",
            "voltage": [0.5, 0.5],
            "span": ["low", "low"],
            "ramping_enabled": [True, False],
            "ramp_rate": [0.01, 0.01],
            "dacs": [10, 11],
            "low_pass_filter": ["dc", "dc"],
        }
    )
    qdac.device = MagicMock()
    return qdac


@pytest.fixture(name="qdac_out_range")
def range_input() -> QDevilQDac2:
    """Fixture that returns an instance of a dummy QDAC-II."""

    qdac_out_range = QDevilQDac2(
        {
            "alias": "qdac_out_range",
            "voltage": [0.5],
            "span": ["low"],
            "ramping_enabled": [True],
            "ramp_rate": [0.001],
            "dacs": [10],
            "low_pass_filter": ["dc"],
        }
    )
    mock_device = MagicMock()
    qdac_out_range.device = mock_device
    return qdac_out_range


@pytest.fixture(name="waveform")
def get_square_waveform() -> Square:
    return Square(0.1, 4)


class TestQDevilQDac2:
    def test_initial_setup(self, qdac: QDevilQDac2):
        """Test initial_setup method"""
        qdac.initial_setup()

        channel_calls = []
        for channel_id in qdac.dacs:
            index = qdac.dacs.index(channel_id)
            channel_calls.append(call(channel_id))
            channel_calls.append(call().dc_mode("fixed"))
            channel_calls.append(call().output_range(qdac.span[index]))
            channel_calls.append(call().output_filter(qdac.low_pass_filter[index]))
            if qdac.ramping_enabled[index]:
                channel_calls.append(call().dc_slew_rate_V_per_s(qdac.ramp_rate[index]))
            else:
                channel_calls.append(call().dc_slew_rate_V_per_s(2e7))
            channel_calls.append(call().dc_constant_V(0.0))

        qdac.device.channel.assert_has_calls(channel_calls)

    def test_turn_on(self, qdac: QDevilQDac2):
        """Test turn_on method"""
        qdac.turn_on()

        channel_calls = []
        for channel_id in qdac.dacs:
            index = qdac.dacs.index(channel_id)
            channel_calls.append(call(channel_id))
            channel_calls.append(call().dc_constant_V(qdac.voltage[index]))

        qdac.device.channel.assert_has_calls(channel_calls)

    def test_turn_off(self, qdac: QDevilQDac2):
        """Test turn_off method"""
        qdac.turn_off()

        channel_calls = []
        for channel_id in qdac.dacs:
            channel_calls.append(call(channel_id))
            channel_calls.append(call().dc_constant_V(0.0))

        qdac.device.channel.assert_has_calls(channel_calls)

    def test_reset(self, qdac: QDevilQDac2):
        """Test reset method"""
        qdac.reset()

        qdac.device.reset.assert_called_once()

    def test_get_dac(self, qdac: QDevilQDac2):
        """Test get_dac method"""
        channel_calls = []
        for channel_id in qdac.dacs:
            qdac.get_dac(channel_id)
            channel_calls.append(call(channel_id))
        qdac.device.channel.assert_has_calls(channel_calls)

    def test_upload_waveform(self, qdac: QDevilQDac2, waveform: Square):
        """Test upload_waveform method"""
        channel_id = 4
        qdac.upload_waveform(waveform, channel_id)
        qdac.device.allocate_trace.assert_called_once_with(channel_id, len(waveform.envelope()))
        assert qdac._cache == {4: True}

    def test_upload_waveform_fails_overwrite_cache(self, qdac: QDevilQDac2, waveform: Square):
        """Test that upload waveform raises an error when trying to allocate a waveform to an already allocated channel id"""
        channel_id = 2
        qdac._cache = {channel_id: True}
        error_string = re.escape(f"Device {qdac.name} already has a waveform allocated to channel {channel_id}. Clear the cache before allocating a new waveform")
        with pytest.raises(ValueError, match=error_string):
            qdac.upload_waveform(waveform, channel_id)

    def test_upload_waveform_fails_odd_value(self, qdac: QDevilQDac2, waveform: Square):
        """Test that upload waveform raises an error when uploading a waveform with odd number of entries"""
        channel_id = 2
        waveform = Square(0.1, 3)
        error_string = "Waveform entries must be even."
        with pytest.raises(ValueError, match=error_string):
            qdac.upload_waveform(waveform, channel_id)

    def test_upload_waveform_fails_amp_range(self, qdac: QDevilQDac2, waveform: Square):
        """Test that upload waveform raises an error when uploading a waveform with outside the allowed amplitude range"""
        channel_id = 2
        waveform = Square(1, 4)
        error_string = re.escape("Waveform amplitudes must be within [-1,1] range.")
        with pytest.raises(ValueError, match=error_string):
            qdac.upload_waveform(waveform, channel_id)
        qdac.clear_cache()
        channel_id = 2
        waveform = Square(-1.1, 4)
        with pytest.raises(ValueError, match=error_string):
            qdac.upload_waveform(waveform, channel_id)

    def test_play(self, qdac: QDevilQDac2):
        """Test play method"""
        channel_id = 4
        channel_calls = [call(10), call(11), call(4), call().start]
        qdac._cache = {channel_id: True}
        qdac.play(clear_after=False)
        # cache not erased if default clear_after
        assert qdac._cache == {channel_id: True}
        qdac.play(channel_id)
        qdac.get_dac(4).arbitrary_wave.assert_has_calls(channel_calls)
        # check that cache is erased
        assert qdac._cache == {}

    def test_clear_cache(self, qdac: QDevilQDac2):
        """Test clear_cache method"""
        qdac._cache = {2: True}
        qdac.clear_cache()
        assert qdac._cache == {}
        qdac.device.remove_traces.assert_called_once()

    def test_input_range_runcard(self, qdac_out_range: QDevilQDac2):
        # Test that an error is raised when the input value on the runcard for the qdac are out of bound
        channel_id = 10
        error_string = re.escape(f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s.")
        with pytest.raises(ValueError, match=error_string):
            qdac_out_range.initial_setup()

    def test_input_range_set_parameter(self, qdac: QDevilQDac2):
        # Test that an error is raised when the input value on set_parameter RAMPING_RATE for the qdac are out of bound
        channel_id = 10
        error_string = re.escape(f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s.")
        with pytest.raises(ValueError, match=error_string):
            qdac.set_parameter(parameter=Parameter.RAMPING_RATE, value=0.0001, channel_id=channel_id)

    def test_input_range_set_parameter_enabled(self, qdac_out_range: QDevilQDac2):
        # Test that an error is raised when the input value on set_parameter RAMPING_ENABLED for the qdac are out of bound
        channel_id = 10
        error_string = re.escape(f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s.")
        with pytest.raises(ValueError, match=error_string):
            qdac_out_range.set_parameter(parameter=Parameter.RAMPING_ENABLED, value=True, channel_id=channel_id)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.VOLTAGE, -0.001),
            (Parameter.RAMPING_ENABLED, False),
            (Parameter.RAMPING_ENABLED, True),
            (Parameter.RAMPING_RATE, 0.05),
            (Parameter.SPAN, "high"),
            (Parameter.LOW_PASS_FILTER, "low"),
        ],
    )
    def test_set_parameter_method(self, qdac: QDevilQDac2, parameter: Parameter, value):
        """Test setup method"""
        for index, channel_id in enumerate(qdac.dacs):
            qdac.set_parameter(parameter=parameter, value=value, channel_id=channel_id)

            channel_calls = []
            channel_calls.append(call(channel_id))
            if parameter == Parameter.VOLTAGE:
                channel_calls.append(call().dc_constant_V(value))
            if parameter == Parameter.SPAN:
                channel_calls.append(call().output_range(value))
            if parameter == Parameter.LOW_PASS_FILTER:
                channel_calls.append(call().output_filter(value))
            if parameter == Parameter.RAMPING_ENABLED:
                if value:
                    channel_calls.append(call().dc_slew_rate_V_per_s(qdac.ramp_rate[index]))
                else:
                    channel_calls.append(call().dc_slew_rate_V_per_s(2e7))
            if parameter == Parameter.RAMPING_RATE:
                if qdac.ramping_enabled[index]:
                    channel_calls.append(call().dc_slew_rate_V_per_s(qdac.ramp_rate[index]))

            qdac.device.channel.assert_has_calls(channel_calls)
            assert qdac.get_parameter(parameter=parameter, channel_id=channel_id) == value

    @pytest.mark.parametrize("parameter, value", [(Parameter.MAX_CURRENT, 0.001), (Parameter.GAIN, 0.0005)])
    def test_set_parameter_method_raises_exception(self, qdac: QDevilQDac2, parameter: Parameter, value):
        """Test the setup method raises an exception with wrong parameters"""
        for channel_id in qdac.dacs:
            with pytest.raises(ParameterNotFound):
                qdac.set_parameter(parameter, value, channel_id)

    @pytest.mark.parametrize("parameter, value", [(Parameter.MAX_CURRENT, 0.001), (Parameter.GAIN, 0.0005)])
    def test_get_parameter_method_raises_exception(self, qdac: QDevilQDac2, parameter: Parameter, value):
        """Test the get method raises an exception with wrong parameters"""
        for channel_id in qdac.dacs:
            with pytest.raises(ParameterNotFound):
                qdac.get_parameter(parameter, channel_id)

    @pytest.mark.parametrize("channel_id", [0, 25, -1, None])
    def test_validate_channel_method_raises_exception(self, qdac: QDevilQDac2, channel_id):
        """Test the _validate_channel method raises an exception with wrong parameters"""
        with pytest.raises(ValueError):
            qdac._validate_channel(channel_id)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code="attr-defined"
from datetime import datetime
from typing import TypedDict

import h5py
import numpy as np

from qililab.result.experiment_results import ExperimentResults


class VariableMetadata(TypedDict):
    """Metadata for a variable used in the experiment.

    Attributes:
        label (str): The label of the variable.
        values (np.ndarray): The array of values for the variable.
    """

    label: str
    values: np.ndarray


class MeasurementMetadata(TypedDict):
    """Metadata for a measurement in the experiment.

    Attributes:
        variables (list[VariableMetadata]): List of variables for the measurement.
        dims (list[list[str]]): Dimensions of the measurement data.
        shape (tuple[int, ...]): Shape of the measurement data.
        shots (int): Number of shots taken for the measurement.
    """

    variables: list[VariableMetadata]
    dims: list[list[str]]
    shape: tuple[int, ...]
    shots: int


class QProgramMetadata(TypedDict):
    """Metadata for a quantum program.

    Attributes:
        variables (dict[str, np.ndarray]): Variables that the QProgram depends upon.
        dims (list[list[str]]): Dimensions of the QProgram data.
        measurements (dict[str, MeasurementMetadata]): Measurements associated with the QProgram.
    """

    variables: dict[str, np.ndarray]
    dims: list[list[str]]
    measurements: dict[str, MeasurementMetadata]


class ExperimentMetadata(TypedDict, total=False):
    """Metadata for an experiment.

    Attributes:
        platform (str): Platform information.
        experiment (str): Experiment description.
        executed_at (datetime): Timestamp when the experiment started execution.
        execution_time (float): Time taken for the execution in seconds.
        qprograms (dict[str, QProgramMetadata]): Quantum programs included in the experiment.
    """

    platform: str
    experiment: str
    executed_at: datetime
    execution_time: float
    qprograms: dict[str, QProgramMetadata]


class ExperimentResultsWriter(ExperimentResults):
    """
    Allows for real-time saving of results from an experiment using the provided metadata information.

    Inherits from `ExperimentResults` to support both read and write operations.
    """

    def __init__(self, path: str, metadata: ExperimentMetadata):
        """Initializes the ExperimentResultsWriter instance.

        Args:
            path (str): The file path to save the HDF5 results file.
            metadata (ExperimentMetadata): The metadata describing the experiment structure.
        """
        super().__init__(path)
        self._metadata = metadata

    # pylint: disable=too-many-locals
    def _create_results_file(self):
        """Creates the HDF5 file structure and registers loops as dimension scales.

        Writes the metadata to the HDF5 file and sets up the datasets and groups for streaming data.
        """
        h5py.get_config().track_order = True

        if "platform" in self._metadata:
            self.platform = self._metadata["platform"]

        if "experiment" in self._metadata:
            self.experiment = self._metadata["experiment"]

        if "executed_at" in self._metadata:
            self.executed_at = self._metadata["executed_at"]

        if "execution_time" in self._metadata:
            self.execution_time = self._metadata["execution_time"]

        if "qprograms" in self._metadata:
            # Create the group for QPrograms
            qprograms_group = self._file.create_group(ExperimentResultsWriter.QPROGRAMS_PATH)

            # Iterate through QPrograms and measurements in the structure
            for qprogram_name, qprogram_data in self._metadata["qprograms"].items():
                qgroup = qprograms_group.create_group(qprogram_name)

                # Write variables that QProgram depends upon (software loops)
                qloop_group = qgroup.create_group(ExperimentResults.VARIABLES_PATH)
                for variable in qprogram_data["variables"]:
                    label, values = variable["label"], variable["values"]
                    loop = qloop_group.create_dataset(label, data=values)
                    loop.dims[0].label = label
                    loop.make_scale(label)  # Register as dimension scale

                # Write measurements
                measurement_group = qgroup.create_group(ExperimentResults.MEASUREMENTS_PATH)
                for measurement_name, measurement_data in qprogram_data["measurements"].items():
                    mgroup = measurement_group.create_group(measurement_name)
                    mloop_group = mgroup.create_group(ExperimentResults.VARIABLES_PATH)

                    # Write shots of measurement
                    mgroup["shots"] = measurement_data["shots"]

                    # Write variables that measurement depends upon (hardware loops)
                    for variable in measurement_data["variables"]:
                        label, values = variable["label"], variable["values"]
                        loop = mloop_group.create_dataset(label, data=values)
                        loop.dims[0].label = label
                        loop.make_scale(label)  # Register as dimension scale

                    # Create the results dataset
                    results_ds = mgroup.create_dataset(ExperimentResults.RESULTS_PATH, shape=measurement_data["shape"])

                    # Attach dimension scales (loops) to the results dataset
                    for idx, dim_variables in enumerate(qprogram_data["dims"]):
                        for dim_variable in dim_variables:
                            results_ds.dims[idx].attach_scale(qloop_group[dim_variable])
                        results_ds.dims[idx].label = ",".join(list(dim_variables))
                    for idx, dim_variables in enumerate(measurement_data["dims"]):
                        for dim_variable in dim_variables:
                            results_ds.dims[len(qprogram_data["dims"]) + idx].attach_scale(mloop_group[dim_variable])
                        results_ds.dims[len(qprogram_data["dims"]) + idx].label = ",".join(list(dim_variables))

                    # Attach the extra dimension (usually for I/Q) to the results dataset
                    results_ds.dims[len(qprogram_data["dims"]) + len(measurement_data["dims"])].label = "I/Q"

    def _create_resuts_access(self):
        """Sets up internal data structures to allow for real-time data writing to the HDF5 file."""
        if "qprograms" in self._metadata:
            for qprogram_name, qprogram_data in self._metadata["qprograms"].items():
                for measurement_name, _ in qprogram_data["measurements"].items():
                    self.data[qprogram_name, measurement_name] = self._file[
                        f"qprograms/{qprogram_name}/measurements/{measurement_name}/results"
                    ]

    def __enter__(self):
        """Opens the HDF5 file and creates the structure for streaming.

        Returns:
            ExperimentResultsWriter: The ExperimentResultsWriter instance.
        """
        self._file = h5py.File(self.path, mode="w")
        self._create_results_file()
        self._create_resuts_access()

        return self

    def __setitem__(self, key: tuple, value: float):
        """Sets an item in the results dataset.

        Args:
            key (tuple): A tuple of (qprogram_name or index, measurement_name or index, *indices).
            value (float): The value to set at the specified indices.
        """
        qprogram_name, measurement_name, *indices = key
        if isinstance(qprogram_name, int):
            qprogram_name = f"QProgram_{qprogram_name}"
        if isinstance(measurement_name, int):
            measurement_name = f"Measurement_{measurement_name}"
        self.data[qprogram_name, measurement_name][tuple(indices)] = value

    @ExperimentResults.platform.setter
    def platform(self, platform: str):
        """Sets the YAML representation of the platform.

        Args:
            platform (str): The YAML string representing the platform.
        """
        path = ExperimentResults.PLATFORM_PATH
        if path in self._file:
            del self._file[path]
        self._file[path] = platform

    @ExperimentResults.experiment.setter
    def experiment(self, experiment: str):
        """Sets the YAML representation of the executed experiment.

        Args:
            experiment (str): The YAML string representing the experiment.
        """
        path = ExperimentResults.EXPERIMENT_PATH
        if path in self._file:
            del self._file[path]
        self._file[path] = experiment

    @ExperimentResults.executed_at.setter
    def executed_at(self, dt: datetime):
        """Sets the timestamp when execution of the experiment started.

        Args:
            dt (datetime): The datetime when execution started.
        """
        path = ExperimentResults.EXECUTED_AT_PATH
        if path in self._file:
            del self._file[path]
        self._file[path] = dt.strftime("%Y-%m-%d %H:%M:%S")

    @ExperimentResults.execution_time.setter
    def execution_time(self, time: float):
        """Sets the execution time in seconds.

        Args:
            time (float): The execution time in seconds.
        """
        path = ExperimentResults.EXECUTION_TIME_PATH
        if path in self._file:
            del self._file[path]
        self._file[path] = str(time)



import numpy as np

from qililab.digital.native_gates import Drag, _GateHandler
from qibo import Circuit, gates


def test_native_gates_drag():
    """Test drag pulse native gate"""

    drag_gate = Drag(0, 0, 0)  # initialize gate

    assert isinstance(drag_gate, Drag)
    assert drag_gate.name == "drag"
    assert len(drag_gate.parameters) == 2
    assert drag_gate.parameter_names == ["theta", "phase"]
    assert drag_gate.trainable is True

    # try different parameters
    rng = np.random.default_rng(seed=42)  # init random number generator
    for _ in range(50):
        theta = rng.random()
        phi = rng.random()
        qubit = rng.integers(0, 100)
        drag_gate = Drag(qubit, theta, phi)  # initialize gate
        assert drag_gate.parameters == (theta, phi)
        assert drag_gate.qubits == qubit

def test_normalize_angle():
    """Test normalize angle."""
    assert _GateHandler.normalize_angle(3 * np.pi) == np.pi
    assert _GateHandler.normalize_angle(2 * np.pi) == 0
    assert _GateHandler.normalize_angle(3/2 * np.pi) == - np.pi/2
    assert _GateHandler.normalize_angle(np.pi) == np.pi
    assert _GateHandler.normalize_angle(np.pi / 2) == np.pi / 2

def test_get_circuit_gates_info():
    """Test get circuit gates."""
    circuit = Circuit(2)
    circuit.add(gates.X(0))
    circuit.add(gates.H(1))

    circuit_gates_info = _GateHandler.get_circuit_gates_info(circuit.queue)

    assert circuit_gates_info == [("X", [0], {}), ("H", [1], {})]

def test_create_gate():
    """Test create gate."""
    gate = _GateHandler.create_gate("X", [0], {})
    assert isinstance(gate, gates.X)
    assert gate.init_args == [0]

def test_create_circuit():
    """Test create circuit."""
    circuit_gates = [("X", [0], {}), ("H", [1], {})]
    circuit_gates = _GateHandler.create_qibo_gates_from_gates_info(circuit_gates)

    assert len(circuit_gates) == 2
    assert [gate.name for gate in circuit_gates] == ["x", "h"]


def test_create_circuit_from_gates():
    """Test create circuit from gates."""
    gates_list = [gates.X(0), gates.H(1)]
    nqubits = 2
    wire_names = [0, 1]

    circuit = _GateHandler.create_circuit_from_gates(gates_list, nqubits, wire_names)

    assert len(circuit.queue) == 2
    assert [gate.name for gate in circuit.queue] == ["x", "h"]
    assert circuit.wire_names == [0, 1]

def test_extract_qubits():
    """Test extract qubits."""
    qubits = _GateHandler.extract_qubits_from_gate_args([0, 1])
    assert qubits == [0, 1]

    qubits = _GateHandler.extract_qubits_from_gate_args(0)
    assert qubits == [0]



import numpy as np
import pytest

from qililab.waveforms import Ramp


@pytest.fixture(name="ramp_wf")
def fixture_square() -> Ramp:
    return Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=100)


class TestRamp:
    def test_init(self, ramp_wf: Ramp):
        """Test __init__ method"""
        assert ramp_wf.from_amplitude == 0.0
        assert ramp_wf.to_amplitude == 1.0
        assert ramp_wf.duration == 100

    def test_envelope_method(self, ramp_wf: Ramp):
        """Test envelope method"""
        expected_envelope = np.linspace(ramp_wf.from_amplitude, ramp_wf.to_amplitude, ramp_wf.duration)
        envelope = ramp_wf.envelope()
        assert np.allclose(envelope, expected_envelope)

    def test_get_duration_method(self, ramp_wf: Ramp):
        """Test get_duration method"""
        assert ramp_wf.get_duration() == 100



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bias tee correction."""

from dataclasses import dataclass

import numpy as np
from scipy import signal

from qililab.typings import PulseDistortionName
from qililab.utils import Factory

from .pulse_distortion import PulseDistortion


@Factory.register
@dataclass(frozen=True, eq=True)
class BiasTeeCorrection(PulseDistortion):
    """Bias tee distortion. Corrects for a bias tee using a linear IIR filter with time constant tau.

    For more info, check `SUPLEMENTAL MATERIAL <https://arxiv.org/abs/1907.04818>`_.

    Args:
        tau_bias_tee (float): Time constant.
        sampling_rate (float, optional): Sampling rate. Defaults to 1.
        norm_factor (float, optional): The manual normalization factor that multiplies the envelope in the apply() method. Defaults to 1 (no effect).
        auto_norm (bool, optional): Whether to automatically normalize the corrected envelope with the original max height in the apply() method.
            (The max height is the furthest number from 0 in the envelope, only checking the real axis/part). Defaults to True.

    Returns:
        PulseDistortion: Distortion to apply to given envelopes in :class:`PulseEvent`.

    Examples:

        Imagine you want to distort a :class:`Rectangular` envelope with a BiasTeeCorrection. You could do:

        >>> from qililab.pulse import Rectangular, BiasTeeCorrection
        >>> envelope = Rectangular().envelope(duration=50, amplitude=1.0)
        >>> distorted_envelope = BiasTeeCorrection(tau_bias_tee=1.3).apply(envelope)

        which would return a distorted envelope with the same real max height as the initial.

        >>> np.max(distorted_envelope) == np.max(envelope)
        True

        .. note::
            You can find more examples in the docstring of the :class:`PulseDistortion` base class.
    """

    name = PulseDistortionName.BIAS_TEE_CORRECTION  # noqa: RUF045
    tau_bias_tee: float  #: Time constant.
    sampling_rate: float = 1.0  #: Sampling rate. Defaults to 1.

    def apply(self, envelope: np.ndarray) -> np.ndarray:
        """Distorts envelopes (originally created to distort square envelopes).

        Corrects for a bias tee using a linear IIR filter with time constant tau.

        If `self.auto_norm` is True (default) normalizes the resulting envelope to have the same real max height than the starting one.
        (the max height is the furthest number from 0, only checking the real axis/part)
        If the corrected envelope is zero everywhere or doesn't have a real part this process is skipped.

        Finally it applies the manual `self.norm_factor` to the result, reducing the full envelope by its magnitude.

        For further details on the normalization implementation see the docstring on :class:`PulseDistortion` base class.

        Args:
            envelope (numpy.ndarray): Array representing the envelope of a pulse for each time step.

        Returns:
            numpy.ndarray: Amplitude of the envelope for each time step.
        """
        # Parameters
        k = 2 * self.tau_bias_tee * self.sampling_rate

        # Coefficients
        a = [1, -1]
        b = [(k + 1) / k, -(k - 1) / k]

        # Filtered signal
        corr_envelope = signal.lfilter(b=b, a=a, x=envelope)
        return self.normalize_envelope(envelope=envelope, corr_envelope=corr_envelope)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from qililab.qprogram.operations.operation import Operation
from qililab.qprogram.variable import Variable
from qililab.waveforms import IQPair, Waveform
from qililab.yaml import yaml


@yaml.register_class
class Play(Operation):
    def __init__(self, bus: str, waveform: Waveform | IQPair, wait_time: int | None = None) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: Waveform | IQPair = waveform
        self.wait_time: int | None = wait_time  # TODO: remove this in clean fix

    def get_waveforms(self) -> tuple[Waveform, Waveform | None]:
        """Get the waveforms.

        Returns:
            tuple[Waveform, Waveform | None]: The waveforms as tuple. The second waveform can be None.
        """
        wf_I: Waveform = self.waveform.I if isinstance(self.waveform, IQPair) else self.waveform
        wf_Q: Waveform | None = self.waveform.Q if isinstance(self.waveform, IQPair) else None
        return wf_I, wf_Q

    def get_waveform_variables(self) -> set[Variable]:
        """Get a set of the variables used in the waveforms, if any.

        Returns:
            set[Variable]: The set of variables used in the waveforms.
        """
        wf_I, wf_Q = self.get_waveforms()
        variables_I = [attribute for attribute in wf_I.__dict__.values() if isinstance(attribute, Variable)]
        variables_Q = (
            [attribute for attribute in wf_Q.__dict__.values() if isinstance(attribute, Variable)] if wf_Q else []
        )
        return set(variables_I + variables_Q)

    def get_variables(self) -> set[Variable]:
        """Get a set of the variables used in operation, if any.

        Returns:
            set[Variable]: The set of variables used in operation.
        """
        return super().get_variables() | self.get_waveform_variables()


@yaml.register_class
class PlayWithCalibratedWaveform(Operation):
    def __init__(self, bus: str, waveform: str, wait_time: int | None = None) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: str = waveform
        self.wait_time: int | None = wait_time



"""Tests for the PulseSequences class."""
import pytest

from qililab.pulse import Gaussian, Pulse, PulseBusSchedule, PulseEvent, PulseSchedule


@pytest.fixture(name="pulse_event")
def fixture_pulse_event() -> PulseEvent:
    """Load PulseEvent.

    Returns:
        PulseEvent: Instance of the PulseEvent class.
    """
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    return PulseEvent(pulse=pulse, start_time=0)


@pytest.fixture(name="pulse_schedule")
def fixture_pulse_schedule() -> PulseSchedule:
    """Return PulseSchedule instance."""
    schedule = PulseSchedule(elements=[
        PulseBusSchedule(bus_alias="drive_q0", timeline=[
            PulseEvent(Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=Gaussian(num_sigmas=4)), start_time=0)
        ])
    ])
    return schedule


class TestPulseSequences:
    """Unit tests checking the PulseSequences attributes and methods"""

    def test_init(self, pulse_schedule: PulseSchedule):
        assert isinstance(pulse_schedule, PulseSchedule)
        assert isinstance(pulse_schedule.elements, list)
        assert len(pulse_schedule.elements) == 1
        assert len(pulse_schedule.elements[0].timeline) == 1

    def test_add_event_method(self, pulse_schedule: PulseSchedule, pulse_event: PulseEvent):
        """Tead add_event method."""
        pulse_event = PulseEvent(pulse=Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=Gaussian(num_sigmas=4)), start_time=100)
        pulse_schedule.add_event(pulse_event=pulse_event, bus_alias="drive_q0", delay=0)
        assert len(pulse_schedule.elements) == 1
        assert len(pulse_schedule.elements[0].timeline) == 2

    def test_to_dict_method(self, pulse_schedule: PulseSchedule):
        """Test to_dict method"""
        dictionary = pulse_schedule.to_dict()
        assert isinstance(dictionary, dict)

    def test_from_dict_method(self, pulse_schedule: PulseSchedule):
        """Test from_dict method"""
        dictionary = pulse_schedule.to_dict()
        pulse_schedule_2 = PulseSchedule.from_dict(dictionary=dictionary)
        assert isinstance(pulse_schedule_2, PulseSchedule)

    def test_iter_method(self, pulse_schedule: PulseSchedule):
        """Test __iter__ method."""
        for pulse_bus_schedule in pulse_schedule:
            assert isinstance(pulse_bus_schedule, PulseBusSchedule)



import numpy as np
import pytest

from qililab.waveforms import DragCorrection, Gaussian, IQPair, Square


@pytest.fixture(name="iq_pair")
def fixture_iq_pair() -> IQPair:
    return IQPair(I=Square(amplitude=0.5, duration=100), Q=Square(amplitude=1.0, duration=100))


class TestIQPair:
    def test_init(self, iq_pair: IQPair):
        """Test __init__ method."""
        assert getattr(iq_pair.I, "amplitude") == 0.5
        assert getattr(iq_pair.I, "duration") == 100
        assert getattr(iq_pair.Q, "amplitude") == 1.0
        assert getattr(iq_pair.Q, "duration") == 100

    def test_get_duration_method(self, iq_pair: IQPair):
        """Test get_duration method."""
        assert iq_pair.get_duration() == 100

    def test_iq_pair_with_different_durations_throws_error(self):
        """Test that waveforms of an IQ pair must have the same duration."""
        with pytest.raises(ValueError, match="Waveforms of an IQ pair must have the same duration."):
            IQPair(I=Square(amplitude=0.5, duration=200), Q=Square(amplitude=1.0, duration=100))

    def test_iq_pair_without_waveform_type_throws_error(self):
        """Test that waveforms of an IQ pair must have Waveform type."""
        with pytest.raises(TypeError, match="Waveform inside IQPair must have Waveform type."):
            IQPair(
                I=IQPair(I=Square(amplitude=0.5, duration=100), Q=Square(amplitude=1.0, duration=100)),
                Q=IQPair(I=Square(amplitude=0.5, duration=100), Q=Square(amplitude=1.0, duration=100)),
            )

    def test_drag_method(self):
        """Test __init__ method"""
        drag = IQPair.DRAG(drag_coefficient=0.4, amplitude=0.7, duration=40, num_sigmas=2)
        gaus = Gaussian(amplitude=0.7, duration=40, num_sigmas=2)
        corr = DragCorrection(drag_coefficient=0.4, waveform=gaus)

        assert isinstance(drag, IQPair)
        assert isinstance(drag.I, Gaussian)
        assert isinstance(drag.Q, DragCorrection)
        assert np.allclose(drag.I.envelope(), gaus.envelope())
        assert np.allclose(drag.Q.envelope(), corr.envelope())



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Automatic-calibration Node module, which works with notebooks as nodes."""

import json
import logging
import os
from datetime import datetime
from io import StringIO
from typing import Any

import numpy as np
import papermill as pm

from qililab.config import logger

logger_output_start = "RAND_INT:47102512880765720413 - OUTPUTS: "


class CalibrationNode:
    """Automatic calibration Node class representing a node in the calibration graph.

    The calibration graph represents a calibration procedure, where each node represents a step of this calibration procedure. **Each of these steps consists of:**

    - **A Jupyter Notebook** to calibrate or check the data with (including its metadata for execution time checks). Such a notebook should contain the following:

        **1) An input parameters cell** (tagged as `parameters`). These parameters are the ones to be overwritten by the ``input_parameters``.

        **2) An experiment/circuit** with its corresponding loops for the sweep given by ``sweep_interval``.

        **3) An analysis procedure**, that plots and fits the obtained data to the expected theoretical behavior, finding the optimal desired parameters.

        **4) An export data cell**, that calls ``export_nb_outputs()`` with the dictionary to retrieve data from the notebook into the calibration workflow.
        This dictionary contains a ``platform_parameters`` list of parameters to set on the platform and might also contain "fidelities", for the fidelities notebooks.

        .. note::

            More information about the notebooks contents and execution, can be found in the examples below.

    - **Inputs to pass to this notebook (optional)**, which might vary for different calls of the same notebook, such:
        ``sweep_interval``, ``input_parameters`` (kwargs).

    .. note::

        If the same notebook is going to be used multiple times for the same ``qubit``, then you need to pass a ``node_distinguisher`` argument, so the :class:`.CalibrationController`
        makes the graph mapping correctly.

    Args:
        nb_path (str): Full notebook path with the folder, nb_name, and ``.ipynb`` extension, written in unix format: `folder/subfolder/.../file.ipynb`.
        qubit_index (int | list[int], optional): Qubit on which this notebook will be executed. Defaults to None.
        node_distinguishier (int | str, optional): Distinguisher for when the same notebook its used multiple times in the same qubit. Mandatory to use in such case, or
            the :class:`.CalibrationController` won't do the graph mapping properly, and the calibration will fail. Defaults to None.
        input_parameters (dict, optional): Kwargs for input parameters to pass and be interpreted by the notebook. Defaults to None.
        sweep_interval (np.ndarray, optional): Array describing the sweep values of the experiment. Defaults to None, which means the one specified in the notebook will be used.

    Examples:

        **Notebook execution:**

        First the key functionality of this class is implemented in the ``run_node()`` method**. The workflow of ``run_node()`` is as follows:

        1. Prepare any input parameters needed for the notebook, including extra parameters defined by the user and essential ones such as the targeted qubit or the sweep intervals.

        2. Create a file with a temporary name. This file will be used to save the execution of the notebook and initially has the following format:

            ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            The ``_dirty`` flag is added to identify executions that are not completed. Since the data we would find such file is ``dirty``, not completed.

        3. Start the execution of the notebook. There are three possible outcomes:

            3.1) The execution succeeds. If the execution succeeds, the execution file is renamed by updating the timestamp and removing the dirty flag:

                ``NameOfTheNode_TimeExecutionEnded.ipynb``

            3.2) The execution is interrupted. If the execution is interrupted, the ``_dirty`` flag remains in the filename, and the program exits:

                ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            3.2) An exception is thrown. This case is not controlled by the user like interruptions. Instead, exceptions are automatically thrown when
            an error is detected. When an execution error is found, the execution file is renamed with the time the error occurred, adding the `_error` flag, and the program exits:

                ``NameOfTheNode_TimeExecutionFoundError_error.ipynb``

            A more detailed explanation of the error is reported and also described inside the notebook (see `papermill documentation
            <https://papermill.readthedocs.io/en/latest/>`_ for more detailed information).

        At the end of this process, you obtain an executed and saved notebook for manual inspection, along with the optimal parameters to set in the runcard
        and the achieved fidelities.

        ----------

        **Practical example:**

        To create two linked nodes, and pass them to a :class:`.CalibrationController`, you need:

        .. code-block:: python

            import numpy as np
            import networkx as nx

            from qililab.calibration import CalibrationController, CalibrationNode, norm_root_mean_sqrt_error

            # GRAPH CREATION AND NODE MAPPING (key = name in graph, value = node object):
            nodes = {}
            G = nx.DiGraph()
            qubit = 0

            # CREATE NODES :
            for qubit in [0, 1]:
                first = CalibrationNode(
                    nb_path="notebooks/first.ipynb",
                    qubit_index=qubit,
                )
                nodes[first.node_id] = first

                second = CalibrationNode(
                    nb_path="notebooks/second.ipynb",
                    qubit_index=qubit,
                    sweep_interval=np.arange(start=0, stop=19, step=1),
                )
                nodes[second.node_id] = second

                # GRAPH BUILDING (1 --> 2):
                G.add_edge(first.node_id, second.node_id)

            # CREATE CALIBRATION CONTROLLER:
            controller = CalibrationController(node_sequence=nodes, calibration_graph=G, runcard=path_runcard)

            ### WORKFLOW TO DO:
            controller.maintain(nodes["second_q1"])  # maintain second node for qubit 1

        .. note::

            You can find the above code, but defining ``first`` and ``second`` as lists, in the :class:`CalibrationController` class documentation.


        |

        where **the notebooks ``first/second.ipynb``, would contain** the following:

        |

        **1) An input parameters cell** (tagged as `parameters`). These parameters are the ones to be overwritten by the ``input_parameters``:

            .. code-block:: python

                import numpy as np

                qubit = 0

                # Sweep interval:
                sweep_interval = np.arange(start=0, stop=19, step=1)

                # Extra parameters for this concrete notebook:
                param1 = 0
                param2 = 0
                ...

        |

        **2) An experiment/circuit** with its corresponding loops for the sweep given by ``sweep_interval``.

            .. code-block:: python

                # Set the environment and paths:
                ...

                # Set the platform:
                platform = ql.build_platform(path=platform_path)
                platform.connect()
                platform.initial_setup()
                platform.turn_on_instruments()

                # Define circuit
                circuit = ...

                # Loop over the sweeps executing the platform:
                results = []
                for X in sweep_interval:
                    platform.set_parameter(alias=alias, parameter=ql.Parameter.X, value=X)
                    result = platform.execute(program=circuit, num_avg=hw_avg, repetition_duration=repetition_duration)
                    results.append(result.array)

                results = np.hstack(results)

        |

        **3) An analysis procedure**, that plots and fits the obtained data to the expected theoretical behaviour and finds the optimal desired parameters.

            .. code-block:: python

                def fit(xdata, results): ...


                fitted_values, x_data, y_data, figure = fit(xdata=sweep_interval, results=results)
                plt.show()

        |

        **4) An export data cell**, that calls ``export_nb_outputs()`` with the dictionary to retrieve from the notebook into the calibration workflow:

            .. code-block:: python

                from qililab.automatic_calibration.calibration_node import export_nb_outputs

                export_nb_outputs(
                    {
                        "platform_parameters": [
                            (param_name0, fitted_values[0], bus_alias0, qubit),
                            (param_name1, fitted_values[1], bus_alias1, qubit),
                        ],
                        "fidelities": [
                            (qubit, "fidelity1", 0.9),
                            (qubit, "fidelity2", 0.95),
                        ],  # Fidelities in the output dictionary are optional.
                    }
                )

        where the ``platform_parameters`` are a list of parameters to set on the platform.

        .. note::

            If more than one `export_nb_outputs()` is used in the notebook, only the last one will be collected.

    """

    def __init__(
        self,
        nb_path: str,
        qubit_index: int | list[int] | None = None,
        node_distinguisher: int | str | None = None,
        input_parameters: dict | None = None,
        sweep_interval: np.ndarray | None = None,
    ):
        if len(nb_path.split("\\")) > 1:
            raise ValueError("`nb_path` must be written in unix format: `folder/subfolder/.../file.ipynb`.")

        if isinstance(qubit_index, list) and len(qubit_index) != 2:
            raise ValueError("List of `qubit_index` only accepts two qubit index")

        self.nb_path: str = os.path.abspath(nb_path)
        """Absolute notebook path, with folder, nb_name and ``.ipynb`` extension."""

        self.qubit_index: int | list[int] | None = qubit_index
        """Qubit which this notebook will be executed on."""

        self.node_distinguisher: int | str | None = node_distinguisher
        """Distinguisher for when the same notebook its used multiple times in the same qubit. Mandatory to use in such case, or
        the :class:`.CalibrationController` won't do the graph mapping properly, and the calibration will fail. Defaults to None.
        """

        self.node_id, self.nb_folder = self._path_to_name_and_folder(self.nb_path)
        """Node name and folder, separated, and without the ``.ipynb`` extension."""

        self.input_parameters: dict | None = input_parameters
        """Kwargs for input parameters, to pass and then be interpreted by the notebook. Defaults to None."""

        self.sweep_interval: np.ndarray | None = sweep_interval
        """Array describing the sweep values of the experiment. Defaults to None, which means the one specified in the notebook will be used."""

        self.output_parameters: dict | None = None
        """Output parameters dictionary from the notebook execution, which was extracted with ``ql.export_nb_outputs()``, normally contains
        a ``platform_parameters`` which will be the calibrated parameters to set in the platform and might also contain "fidelities", for
        the fidelities notebooks. If no previous successful calibration, then is None.
        """

        self.previous_timestamp: float | None = self.get_last_calibrated_timestamp()
        """Last calibrated timestamp. If no previous successful calibration, then is None."""

        self._stream: StringIO = self._build_notebooks_logger_stream()
        """Stream object to which the notebooks logger output will be written, to posterior retrieval."""

        self.been_calibrated: bool = False
        """Flag whether this notebook has been already calibrated in a concrete run. Defaults to False."""

    def run_node(self) -> float:
        """Executes the notebook, passing the needed parameters and flags.

        **Its workflow is the following:**

        1. Prepare any input parameters needed for the notebook. This includes extra parameters defined by the user and essential ones such as the targeted qubit or the sweep intervals.

        2. Create a file with a temporary name. This file will be used to save the execution of the notebook, and at this stage has the following format:

            ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            The "_dirty" flag is added to the execution files in order to identify the executions that are not completed. So we know that the data we find
            if we open that file is "dirty", not completed.

        3. Start execution of the notebook. From where we can expect 3 possible outcomes from the execution:

            3.1) The execution succeds. If the execution succeds, we rename the execution file by updating the timestamp and removing the dirty flag:

                ``NameOfTheNode_TimeExecutionEnded.ipynb``

            3.2) The execution is interrupted. If the execution is interrupted, the "_dirty" flag will remain in the filename for ever. Notice after an
            interruption the program exits.

                ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            3.2) An exception is thrown. This case is not controlled by the user like the interruptions, instead those exceptions are automatically
            thrown when an error is detected. When an execution error is found, the execution file is renamed with the time that the error ocurred and adding the flag "_error":

                ``NameOfTheNode_TimeExecutionFoundError_error.ipynb``

            A more detailed explanation of the error is reported and also described inside the notebook (see `papermill documentation
            <https://papermill.readthedocs.io/en/latest/>`_ for more detailed information).Then after we post-processed the file, the program exits.

        at the end, from all of this, you will obtain, a executed and saved notebook to manually check, and an outputs dictionary, containing the optimal
        parameters to set in the runcard, together with the achieved fidelities.

        Returns:
            float: Timestamp to identify the notebook execution moment.

        Exits:
            In case of a keyboard interruption or any exception during the execution of the notebook.
        """
        # Create the input parameters for the notebook:)
        params: dict = {}

        if isinstance(self.qubit_index, int):
            params |= {"qubit": self.qubit_index}
        elif isinstance(self.qubit_index, list):
            params |= {"control_qubit": self.qubit_index[0], "target_qubit": self.qubit_index[1]}

        if self.sweep_interval is not None:
            params["sweep_interval"] = self.sweep_interval

        if self.input_parameters is not None:
            params |= self.input_parameters

        # JSON serialize nb input, no np.ndarrays
        params = _json_serialize(params)

        # initially the file is "dirty" until we make sure the execution was not aborted, so we add _dirty tag.
        output_path = self._create_notebook_datetime_path(dirty=True)

        # Execute notebook without problems:
        try:
            self.output_parameters = self._execute_notebook(self.nb_path, output_path, params)
            timestamp = datetime.timestamp(datetime.now())
            # remove the _dirty tag, since it finished.
            new_output_path = self._create_notebook_datetime_path(timestamp=timestamp)
            os.rename(output_path, new_output_path)
            return timestamp

        # When keyboard interrupt (Ctrl+C), generate error, and leave `_dirty` in the name:
        except KeyboardInterrupt as exc:  # we don't remove the _dirty tag, since it was stopped, not failed.
            logger.error("Interrupted automatic calibration notebook execution of %s", self.nb_path)
            raise KeyboardInterrupt(f"Interrupted automatic calibration notebook execution of {self.nb_path}") from exc

        # When notebook execution fails, generate error folder and move there the notebook:
        except Exception as exc:
            if output_path in [entry.path for entry in os.scandir(os.getcwd())]:
                timestamp = datetime.timestamp(datetime.now())
                error_path = self._create_notebook_datetime_path(
                    timestamp=timestamp, error=True
                )  # add _error tag, for failed executions.
                os.rename(output_path, error_path)
                logger.error(
                    "Aborting execution. Exception %s during automatic calibration notebook execution, trace of the error can be found in %s",
                    str(exc),
                    error_path,
                )

                raise Exception(
                    f"Aborting execution. Exception {exc!s} during automatic calibration notebook execution, trace of the error can be found in {error_path}"
                ) from exc

            logger.error(
                "Aborting execution. Exception %s during automatic calibration, expected error execution file to be created but it did not",
                str(exc),
            )

            raise Exception(
                f"Aborting execution. Exception {exc!s} during automatic calibration, expected error execution file to be created but it did not"
            ) from exc

    @staticmethod
    def _build_notebooks_logger_stream() -> StringIO:
        """Build a stream object for capturing the notebook's log output.

        Returns:
            StringIO: Stream object for capturing execution outputs.
        """
        stream = StringIO()
        logging.basicConfig(stream=stream, level=logging.INFO)

        return stream

    def _execute_notebook(self, input_path: str, output_path: str, parameters: dict | None = None) -> dict | None:
        """Executes a Jupyter Notebook overwriting the `parameters` cell, and capturing the execution ``output``.

        This method changes the working directory to the notebook folder before executing the notebook and restores the original working directory after execution if necessary.

        Args:
            input_path (str): The input path of the notebook to be executed.
            output_path (str): The output path where the executed noteboo will be saved. If None, no file will be saved.
            parameters (dict, optional): Input parameters kwargs, to overwrite the notebook `parameters` cell with. Defaults to None.

        Returns:
            dict | None: A dictionary containing the output parameters of the execution.

        Raises:
            IncorrectCalibrationOutput: In case no outputs or multiple outputs where found.
        """
        # Save previous working directory and setup notebook folder as working directory
        original_wd = os.getcwd()
        os.chdir(self.nb_folder)
        pm.execute_notebook(input_path, output_path, parameters, log_output=True, stdout_file=self._stream)
        # Restore previous working directory after execution is done
        os.chdir(original_wd)

        # Retrieve the logger info and extract the output from it:
        logger_string = self._stream.getvalue()
        return self._from_logger_string_to_output_dict(logger_string, input_path)

    def _create_notebook_datetime_path(
        self, timestamp: float | None = None, dirty: bool = False, error: bool = False
    ) -> str:
        """Create a timestamped notebook path, adding the datetime to the file name end, just before the ``.ipynb``.

        If the path directory doesn't exist, it gets created.

        The passed path can have the ``.ipynb`` extension or not.

        The part of the string after the last "/" will be considered the file name, and the part before its directory.

        Args:
            original_path (str): The original path of the notebook, to add the datetime to. The path directory doesn't need to exist. Can have the ``.ipynb`` extension or not.
                The part of the string after the last "/" will be considered the file name, and the part before it's directory.
            timestamp (float, optional): Timestamp to add to the name. If None, the current time will be used. Defaults to None.
            dirty (bool, optional): Flag indicating if the notebook is in a "dirty" state. Defaults to False.
            error (bool, optional): Flag indicating if the notebook comes from an execution error. Defaults to False.

        Returns:
            str: The timestamped notebook absolute path.
        """
        # Create datetime pathHM
        now = datetime.now() if timestamp is None else datetime.fromtimestamp(timestamp)
        daily_path = f"{now.year}_{now.month:02d}_{now.day:02d}"
        now_path = f"{daily_path}-" + f"{now.hour:02d}:{now.minute:02d}:{now.second:02d}"

        # If doesn't exist, create the needed folder for the path
        os.makedirs(self.nb_folder, exist_ok=True)

        # Path of the execution
        flag = "_error" if error else ("_dirty" if dirty else "")

        # Return the string where saved
        return f"{self.nb_folder}/{self.node_id}_{now_path}{flag}.ipynb"

    def _path_to_name_and_folder(self, original_path: str) -> tuple[str, str]:
        """Extract the name and folder from a notebook path. Name will be extended with the qubit it acts on.

        The passed path can have the ``.ipynb`` extension or not.

        The part of the string after the last "/" will be considered the file name, and the part before its directory.

        Args:
            original_path (str): The original path of the notebook. Can have the ``.ipynb`` extension or not.
                The part of the string after the last "/" will be considered the file name, and the part before it's directory.

        Returns:
            tuple[str, str]: A tuple containing the notebook name and its folder.
        """
        # Create qubit_string to add:
        qubit_str = (
            f"_q{self.qubit_index!s}"
            if isinstance(self.qubit_index, int)
            else "_" + "".join(f"q{q}" for q in self.qubit_index)
            if isinstance(self.qubit_index, list)
            else ""
        )

        # Create distinguish_string to differentiate multiple calls of the same node:
        distinguish_str = f"_{self.node_distinguisher!s}" if self.node_distinguisher is not None else ""

        # Remove .ipynb from end if it has one, and separate the folder and name with the last "/":
        split_path = original_path.split(".ipynb")[0].split("/")

        name = split_path.pop() + distinguish_str + qubit_str
        folder_path = "/".join(split_path)
        return name, folder_path

    def get_last_calibrated_timestamp(self) -> float | None:
        """Gets the timestamp of the last successful calibration of the notebook.

        Searches the directory for self.node_id+“_calibrated” and gets the latest creation time.

        Returns:
            float | None: The timestamp of the last successful calibration, or None if no calibration has been completed.
        """
        last_modified_file_name = self._find_last_executed_calibration()
        return (
            os.path.getmtime(os.path.join(self.nb_folder, last_modified_file_name))
            if last_modified_file_name is not None
            else None
        )

    def _find_last_executed_calibration(self) -> str | None:
        """Finds the last executed calibration of the node.

        Returns:
            str | None: The path of the last calibrated notebook, or None if none has been executed.
        """
        # Filtering by node_id and calibrations
        file_names = [
            entry.name
            for entry in os.scandir(self.nb_folder)
            if entry.is_file()
            and f"{self.node_id}" in entry.name
            and "_calibrated" in entry.name.split(f"{self.node_id}")[1]
        ]
        # Get the last created file, most recent one
        return (
            max(file_names, key=lambda fname: os.path.getctime(os.path.join(self.nb_folder, fname)))
            if file_names != []
            else None
        )

    def _from_logger_string_to_output_dict(self, logger_string: str, input_path: str) -> dict | None:
        """Returns the output dictionary from a logger output string. Raises errors if the output doesn't follow the expected format.

        Args:
            logger_string (str): The logger string containing the output dictionary to extract.
            input_path (str): Path of the notebook that generated the output, to raise errors.

        Returns:
            dict | None: The output dictionary of the file execution.

        Raises:
            IncorrectCalibrationOutput: In case multiple outputs where found.
        """
        logger_splitted = logger_string.split(logger_output_start)
        # In case no output is found we return None:
        if len(logger_splitted) < 2:
            return None
        # In case more than one output is found, we keep the last one, and raise a warning:
        if len(logger_splitted) > 2:
            logger.warning("If you had multiple outputs exported in %s, the last one found will be used.", input_path)

        # This next line is for taking into account other encodings, where special characters get `\\` in front.
        clean_data = logger_splitted[-1].split("\\n")[0].replace('\\"', '"')

        logger_outputs_string = clean_data.split("/n")[0].split("\n")[0]  # Remove the last \n for both bars directions
        return json.loads(logger_outputs_string)  # in-dictionary strings will need to be double-quoted "" not ''.

    def _add_string_to_checked_nb_name(self, string_to_add: str, timestamp: float) -> None:
        """Adds a string to the checked notebook name.

        Args:
            string_to_add (str): The string to append to the notebook name.
            timestamp (float): The timestamp to use for the new notebook execution name.
        """
        timestamp_path = self._create_notebook_datetime_path(timestamp=timestamp).split(".ipynb")[0]

        os.rename(f"{timestamp_path}.ipynb", f"{timestamp_path}_{string_to_add}.ipynb")


def export_nb_outputs(outputs: dict) -> None:
    """Function to export notebook outputs into a stream, later collected by the :class:`CalibrationNode` class.

    If this function is used more than once in a notebook, only the last one will be collected.

    Args:
        outputs (dict): Outputs from the notebook to export into the automatic calibration workflow.
    """
    outputs = _json_serialize(outputs)
    print(f"{logger_output_start}{json.dumps(outputs)}")  # noqa: T201


def _json_serialize(obj: Any):
    """Function to JSON serialize the input argument.

    Needed to handle input/output of notebook executions from the :class:`CalibrationNode` class.
    This method only looks for np.ndarrays objects to JSON serialize. Any other non-JSON serializable won't be serialized.

    Args:
        obj (Any): Object to serialize
    """
    if isinstance(obj, (list, tuple, np.ndarray)):
        return [_json_serialize(elem) for elem in obj]

    if isinstance(obj, dict):
        return {_json_serialize(k): _json_serialize(v) for k, v in obj.items()}

    if isinstance(obj, np.integer):
        return int(obj)

    if isinstance(obj, np.floating):
        return float(obj)

    if isinstance(obj, (complex, np.complex128, np.complex64, np.complex128)):
        return {"real": float(obj.real), "imag": float(obj.imag)}

    if isinstance(obj, np.bool_):
        return bool(obj)

    if isinstance(obj, np.void):
        return None

    return obj



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains all the variables used inside a QProgram."""

from enum import Enum
from uuid import UUID, uuid4

from qililab.yaml import yaml


@yaml.register_class
class Domain(Enum):
    """Domain class."""

    Scalar = 0
    Time = 1
    Frequency = 2
    Phase = 3
    Voltage = 4
    Flux = 5

    @classmethod
    def to_yaml(cls, representer, node):
        """Method to be called automatically during YAML serialization."""
        return representer.represent_scalar("!Domain", f"{node.name}-{node.value}")

    @classmethod
    def from_yaml(cls, _, node):
        """Method to be called automatically during YAML deserialization."""
        _, value = node.value.split("-")
        value = int(value)
        return cls(value)


@yaml.register_class
class Variable:
    """Variable class used to define variables inside a QProgram."""

    def __init__(self, label: str, domain: Domain = Domain.Scalar) -> None:
        self._uuid: UUID = uuid4()
        self._label: str = label
        self._domain: Domain = domain

    def __repr__(self):
        return f"Variable(uuid={self.uuid!r}, label={self.label}, domain={self.domain})"

    def __hash__(self):
        return hash(self._uuid)

    def __eq__(self, other):
        return other is not None and isinstance(other, Variable) and self._uuid == other._uuid

    @property
    def uuid(self):
        """Get the uuid of the variable

        Returns:
            UUID: The uuid of the variable
        """
        return self._uuid

    @property
    def label(self):
        """Get the label of the variable

        Returns:
            str: The label of the variable
        """
        return self._label

    @property
    def domain(self):
        """Get the domain of the variable

        Returns:
            Domain: The domain of the variable
        """
        return self._domain


@yaml.register_class
class IntVariable(Variable, int):  # type: ignore
    """Integer variable. This class is used to define a variable of type int, such that Python recognizes this class
    as an integer."""

    def __new__(cls, _: str = "", __: Domain = Domain.Scalar):
        # Create a new float instance
        instance = int.__new__(cls, 0)
        return instance

    def __init__(self, label: str = "", domain: Domain = Domain.Scalar):
        Variable.__init__(self, label, domain)


@yaml.register_class
class FloatVariable(Variable, float):  # type: ignore
    """Float variable. This class is used to define a variable of type float, such that Python recognizes this class
    as a float."""

    def __new__(cls, _: str = "", __: Domain = Domain.Scalar):
        # Create a new int instance
        instance = float.__new__(cls, 0.0)
        return instance

    def __init__(self, label: str = "", domain: Domain = Domain.Scalar):
        Variable.__init__(self, label, domain)
